# This file is part of python-ly, https://pypi.python.org/pypi/python-ly
#
# Copyright (c) 2008 - 2015 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
Uses xml.etree to create a Music XML document.

Example::

    musxml = create_musicxml.CreateMusicXML()
    musxml.create_part()
    musxml.create_measure(divs=1)
    musxml.new_note('C', 4, 'whole', 4)
    xml = musxml.musicxml()
    xml.write(filename)

"""

from __future__ import unicode_literals
from __future__ import division

import sys
try:
    import xml.etree.cElementTree as etree
except ImportError:
    import xml.etree.ElementTree as etree


import ly.pkginfo

from .midi_sound_map import midi_sound_map

class CreateMusicXML():
    """ Creates the XML nodes according to the Music XML standard."""

    def __init__(self):
        """Creates the basic structure of the XML without any music."""
        self.root = etree.Element("score-partwise", version="3.0")
        self.tree = etree.ElementTree(self.root)
        self.score_info = etree.SubElement(self.root, "identification")
        encoding = etree.SubElement(self.score_info, "encoding")
        software = etree.SubElement(encoding, "software")
        software.text = "python-ly " + ly.pkginfo.version
        encoding_date = etree.SubElement(encoding, "encoding-date")
        import datetime
        encoding_date.text = str(datetime.date.today())
        self.partlist = etree.SubElement(self.root, "part-list")
        self.part_count = 1

    ##
    # Building the basic Elements
    ##

    def create_title(self, title):
        """Create score title."""
        mov_title = etree.Element("movement-title")
        mov_title.text = title
        self.root.insert(0, mov_title)

    def create_score_info(self, tag, info, attr={}):
        """Create score info."""
        info_node = etree.SubElement(self.score_info, tag, attr)
        info_node.text = info

    def create_partgroup(self, gr_type, num, name=None, abbr=None, symbol=None):
        """Create a new part group."""
        attr_dict = {"type": gr_type, "number": str(num)}
        partgroup = etree.SubElement(self.partlist, "part-group", attr_dict)
        if name:
            group_name = etree.SubElement(partgroup, "group-name")
            group_name.text = name
        if abbr:
            group_abbr = etree.SubElement(partgroup, "group-abbreviation")
            group_abbr.text = abbr
        if symbol:
            group_symbol = etree.SubElement(partgroup, "group-symbol")
            group_symbol.text = symbol

    def create_part(self, name="unnamed", abbr=False, midi=False):
        """Create a new part """
        strnr = str(self.part_count)
        part = etree.SubElement(self.partlist, "score-part", id="P"+strnr)
        partname = etree.SubElement(part, "part-name")
        partname.text = name
        if abbr:
            partabbr = etree.SubElement(part, "part-abbreviation")
            partabbr.text = abbr
        if midi:
            scoreinstr = etree.SubElement(part, "score-instrument", id="P"+strnr+"-I"+strnr)
            instrname = etree.SubElement(scoreinstr, "instrument-name")
            instrname.text = midi

            if midi in midi_sound_map and midi_sound_map[midi]:
                instrsound = etree.SubElement(scoreinstr, "instrument-sound")
                instrsound.text = midi_sound_map[midi]

            midiinstr = etree.SubElement(part, "midi-instrument", id="P"+strnr+"-I"+strnr)
            midich = etree.SubElement(midiinstr, "midi-channel")
            midich.text = strnr
            midiname = etree.SubElement(midiinstr, "midi-name")
            midiname.text = midi
        self.current_part = etree.SubElement(self.root, "part", id="P"+strnr)
        self.part_count += 1
        self.bar_nr = 1

    def create_measure(self, pickup = False, **bar_attrs):
        """Create new measure """
        if pickup and self.bar_nr == 1:
            self.bar_nr = 0
        self.current_bar = etree.SubElement(self.current_part, "measure", number=str(self.bar_nr))
        self.bar_nr +=1
        if bar_attrs:
            self.new_bar_attr(**bar_attrs)

    ##
    # High-level node creation
    ##

    def new_note(self, step, octave, durtype, divdur, alter=0,
                 acc_token=0, voice=1, dot=0, chord=0, grace=(0, 0), stem_dir=0):
        """Create all nodes needed for a normal note. """
        self.create_note()
        if grace[0]:
            self.add_grace(grace[1])
        if chord:
            self.add_chord()
        self.add_pitch(step, alter, octave)
        if not grace[0]:
            self.add_div_duration(divdur)
        self.add_voice(voice)
        self.add_duration_type(durtype)
        if dot:
            for i in range(dot):
                self.add_dot()
        if alter or acc_token:
            if acc_token == '!': # cautionary
                self.add_accidental(alter, caut=True)
            elif acc_token == '?': # parentheses
                self.add_accidental(alter, parenth=True)
            else:
                self.add_accidental(alter)
        if stem_dir:
            self.set_stem_dir(stem_dir)

    def new_unpitched_note(self, step, octave, durtype, divdur, voice=1,
                            dot=0, chord=0, grace=(0, 0)):
        """Create all nodes needed for an unpitched note. """
        self.create_note()
        if grace[0]:
            self.add_grace(grace[1])
        if chord:
            self.add_chord()
        self.add_unpitched(step, octave)
        if not grace[0]:
            self.add_div_duration(divdur)
        self.add_duration_type(durtype)
        self.add_voice(voice)
        if dot:
            for i in range(dot):
                self.add_dot()

    def tuplet_note(self, fraction, bs, ttype, nr, divs, atyp='', ntyp=''):
        """Convert current note to tuplet """
        base = self.mult * bs[0]
        scaling = bs[1]
        a = divs*4*fraction[1]
        b = (1/base)*fraction[0]
        duration = (a/b)*scaling
        self.change_div_duration(duration)
        from fractions import Fraction
        self.mult = Fraction(fraction[1], fraction[0])
        timemod_node = self.get_time_modify()
        if timemod_node:
            self.adjust_time_modify(timemod_node, fraction)
        else:
            self.add_time_modify(fraction)
        if ttype:
            self.add_notations()
            if atyp and ttype != "stop":
                self.add_tuplet_type(nr, ttype, fraction[0], atyp, fraction[1], ntyp)
            else:
                self.add_tuplet_type(nr, ttype)

    def tie_note(self, tie_type):
        self.add_tie(tie_type)
        self.add_notations()
        self.add_tied(tie_type)

    def new_rest(self, duration, durtype, pos, dot, voice):
        """Create all nodes needed for a rest. """
        self.create_note()
        if pos:
            self.add_rest_w_pos(pos[0], pos[1])
        else:
            self.add_rest()
        self.add_div_duration(duration)
        self.add_voice(voice)
        if durtype:
            self.add_duration_type(durtype)
        if dot:
            for i in range(dot):
                self.add_dot()

    def new_articulation(self, artic):
        """Add specified articulation. """
        self.add_notations()
        self.add_articulations()
        self.add_named_artic(artic)

    def new_simple_ornament(self, ornament):
        """Add specified ornament. """
        self.add_notations()
        self.add_ornaments()
        func_call = getattr(self, 'add_'+ornament)
        func_call()

    def new_adv_ornament(self, ornament, args):
        """Add more complex ornament."""
        self.add_notations()
        self.add_ornaments()
        if ornament == "wavy-line":
            self.add_wavyline(args['type'])

    def new_bar_attr(self, clef=0, mustime=0, key=None, mode=0, divs=0, multirest=0):
        """Create all bar attributes set. """
        self.create_bar_attr()
        if divs:
            self.add_divisions(divs)
        if key is not None:
            self.add_key(key, mode)
        if mustime:
            self.add_time(mustime)
        if clef:
            sign, line, octch = clef
            self.add_clef(sign, line, oct_ch=octch)
        if multirest:
            self.add_bar_style(multirest)

    def create_tempo(self, words, metronome, sound, dots):
        self.add_direction()
        if words:
            self.add_dirwords(words)
        if metronome:
            self.add_metron_dir(metronome[0], metronome[1], dots)
            self.add_sound_dir(sound)

    def create_new_node(self, parentnode, nodename, txt):
        """ The Music XML language is extensive.
        This function can be used to create
        a non basic node not covered elsewhere in this script.

        TODO: add attributes
        """
        new_node = etree.SubElement(parentnode, nodename)
        new_node.text = str(txt)


    ##
    # Low-level node creation
    ##

    def add_creator(self, creator, name):
        """Add creator to score info."""
        attr = {'type': creator }
        self.create_score_info("creator", name, attr)

    def add_rights(self, rights, type=None):
        """Add rights to score info."""
        attr = {'type': type} if type else {}
        self.create_score_info("rights", rights, attr)

    def create_note(self):
        """Create new note."""
        self.current_note = etree.SubElement(self.current_bar, "note")
        self.current_notation = None
        self.current_artic = None
        self.current_ornament = None
        self.current_tech = None

    def add_pitch(self, step, alter, octave):
        """Create new pitch."""
        pitch = etree.SubElement(self.current_note, "pitch")
        stepnode = etree.SubElement(pitch, "step")
        stepnode.text = str(step)
        if alter:
            altnode = etree.SubElement(pitch, "alter")
            altnode.text = str(alter)
        octnode = etree.SubElement(pitch, "octave")
        octnode.text = str(octave)

    def add_unpitched(self, step, octave):
        """Create unpitched."""
        unpitched = etree.SubElement(self.current_note, "unpitched")
        stepnode = etree.SubElement(unpitched, "display-step")
        stepnode.text = str(step)
        octnode = etree.SubElement(unpitched, "display-octave")
        octnode.text = str(octave)

    def add_accidental(self, alter, caut=False, parenth=False):
        """Create accidental."""
        attrib = {}
        if caut:
            attrib['cautionary'] = 'yes'
        if parenth:
            attrib['parentheses'] = 'yes'
        acc = etree.SubElement(self.current_note, "accidental", attrib)
        acc_dict = {
        0: 'natural',
        1: 'sharp', -1: 'flat',
        2: 'sharp-sharp', -2: 'flat-flat',
        0.5: 'natural-up', -0.5: 'natural-down',
        1.5: 'sharp-up', -1.5: 'flat-down'
        }
        acc.text = acc_dict[alter]

    def set_stem_dir(self, dir):
        stem_dir = etree.SubElement(self.current_note, "stem")
        stem_dir.text = dir

    def add_rest(self):
        """Create rest."""
        etree.SubElement(self.current_note, "rest")

    def add_rest_w_pos(self, step, octave):
        """Create rest with display position."""
        restnode = etree.SubElement(self.current_note, "rest")
        stepnode = etree.SubElement(restnode, "display-step")
        octnode = etree.SubElement(restnode, "display-octave")
        stepnode.text = str(step)
        octnode.text = str(octave)

    def add_skip(self, duration, forward=True):
        if forward:
            skip = etree.SubElement(self.current_bar, "forward")
        else:
            skip = etree.SubElement(self.current_bar, "backward")
        dura_node = etree.SubElement(skip, "duration")
        dura_node.text = str(duration)

    def add_div_duration(self, divdur):
        """Create new duration """
        self.duration = etree.SubElement(self.current_note, "duration")
        self.duration.text = str(divdur)
        self.mult = 1

    def change_div_duration(self, newdura):
        """Set new duration when tuplet """
        self.duration.text = str(newdura)

    def add_duration_type(self, durtype):
        """Create new type """
        typenode = etree.SubElement(self.current_note, "type")
        typenode.text = str(durtype)

    def add_dot(self):
        """Create a dot """
        etree.SubElement(self.current_note, "dot")

    def add_beam(self, nr, beam_type):
        """Add beam. """
        beam_node = etree.SubElement(self.current_notation, "beam", number=str(nr))
        beam_node.text = beam_type

    def add_tie(self, tie_type):
        """Create node tie (used for sound of tie) """
        # A tie must be directly after a duration
        insert_at = get_tag_index(self.current_note, "duration") + 1
        tie_element = etree.Element("tie", type=tie_type)
        self.current_note.insert(insert_at, tie_element)

    def add_grace(self, slash):
        """Create grace node """
        if slash:
            etree.SubElement(self.current_note, "grace", slash="yes")
        else:
            etree.SubElement(self.current_note, "grace")

    def add_notations(self):
        if self.current_notation is None:
            self.current_notation = etree.SubElement(self.current_note, "notations")

    def add_tied(self, tie_type):
        """Create node tied (used for notation of tie) """
        etree.SubElement(self.current_notation, "tied", type=tie_type)

    def add_time_modify(self, fraction):
        """Create time modification """
        index = get_tag_index(self.current_note, "accidental")
        if index == -1:
            index = get_tag_index(self.current_note, "dot")
        if index == -1:
            index = get_tag_index(self.current_note, "type")
        timemod_node = etree.Element("time-modification")
        actual_notes = etree.SubElement(timemod_node, "actual-notes")
        actual_notes.text = str(fraction[0])
        norm_notes = etree.SubElement(timemod_node, "normal-notes")
        norm_notes.text = str(fraction[1])
        self.current_note.insert(index + 1, timemod_node)

    def get_time_modify(self):
        """Check if time-modification node already exists."""
        return self.current_note.find("time-modification")

    def adjust_time_modify(self, timemod_node, fraction):
        """Adjust existing time-modification node."""
        actual_notes = timemod_node.find("actual-notes")
        an = int(actual_notes.text) * fraction[0]
        actual_notes.text = str(an)
        norm_notes = timemod_node.find("normal-notes")
        nn = int(norm_notes.text) * fraction[1]
        norm_notes.text = str(nn)

    def add_tuplet_type(self, nr, ttype, actnr=0, acttype='', normnr=0, normtype=''):
        """Create tuplet with type attribute """
        tuplnode = etree.SubElement(self.current_notation, "tuplet",
                                    {'number': str(nr), 'type': ttype })
        if actnr:
            actnode = etree.SubElement(tuplnode, "tuplet-actual")
            atn = etree.SubElement(actnode, "tuplet-number")
            atn.text = str(actnr)
            att = etree.SubElement(actnode, "tuplet-type")
            if not acttype:
                acttype = self.current_note.find("type").text
            att.text = acttype
        if normnr:
            normnode = etree.SubElement(tuplnode, "tuplet-normal")
            ntn = etree.SubElement(normnode, "tuplet-number")
            ntn.text = str(normnr)
            ntt = etree.SubElement(normnode, "tuplet-type")
            if not normtype:
                normtype = self.current_note.find("type").text
            ntt.text = normtype

    def add_slur(self, nr, sl_type):
        """Add slur. """
        self.add_notations()
        etree.SubElement(self.current_notation, "slur", {'number': str(nr), 'type': sl_type })

    def add_named_notation(self, notate):
        """Fermata, etc. """
        self.add_notations()
        etree.SubElement(self.current_notation, notate)

    def add_articulations(self):
        """Common function for creating all types of articulations. """
        if self.current_artic is None:
            self.current_artic = etree.SubElement(self.current_notation, "articulations")

    def add_named_artic(self, artic):
        """Add articulation with specified name. """
        etree.SubElement(self.current_artic, artic)

    def add_ornaments(self):
        if self.current_ornament is None:
            self.add_notations()
            self.current_ornament = etree.SubElement(self.current_notation, "ornaments")

    def add_tremolo(self, trem_type, lines):
        self.add_ornaments()
        trem_node = etree.SubElement(self.current_ornament, "tremolo", type=trem_type)
        trem_node.text = str(lines)

    def add_trill(self):
        etree.SubElement(self.current_ornament, "trill-mark")

    def add_turn(self):
        etree.SubElement(self.current_ornament, "turn")

    def add_mordent(self):
        etree.SubElement(self.current_ornament, "mordent")

    def add_prall(self):
        etree.SubElement(self.current_ornament, "inverted-mordent")

    def add_wavyline(self, end_type):
        self.add_ornaments
        etree.SubElement(self.current_ornament, "wavy-line", type=end_type)

    def add_gliss(self, linetype, endtype, nr):
        nodedict = { "line-type": linetype, "number": str(nr), "type": endtype }
        self.add_notations()
        etree.SubElement(self.current_notation, "glissando", nodedict)

    def add_technical(self):
        if self.current_tech is None:
            self.add_notations()
            self.current_tech = etree.SubElement(self.current_notation, "technical")

    def add_fingering(self, finger_nr):
        self.add_technical()
        fing_node = etree.SubElement(self.current_tech, "fingering")
        fing_node.text = str(finger_nr)

    def create_bar_attr(self):
        """Create node attributes """
        self.bar_attr = etree.SubElement(self.current_bar, "attributes")

    def add_divisions(self, div):
        division = etree.SubElement(self.bar_attr, "divisions")
        division.text = str(div)

    def add_key(self, key, mode):
        keynode = etree.SubElement(self.bar_attr, "key")
        fifths = etree.SubElement(keynode, "fifths")
        fifths.text = str(key)
        modenode = etree.SubElement(keynode, "mode")
        modenode.text = str(mode)

    def add_time(self, timesign):
        if len(timesign)==3:
            timenode = etree.SubElement(self.bar_attr, "time", symbol=timesign[2])
        else:
            timenode = etree.SubElement(self.bar_attr, "time")
        beatnode = etree.SubElement(timenode, "beats")
        beatnode.text = str(timesign[0])
        typenode = etree.SubElement(timenode, "beat-type")
        typenode.text = str(timesign[1])

    def add_clef(self, sign, line, nr=0, oct_ch=0):
        if nr:
            clefnode = etree.SubElement(self.bar_attr, "clef", number=str(nr))
        else:
            clefnode = etree.SubElement(self.bar_attr, "clef")
        signnode = etree.SubElement(clefnode, "sign")
        signnode.text = str(sign)
        if line:
            linenode = etree.SubElement(clefnode, "line")
            linenode.text = str(line)
        if oct_ch:
            octchnode = etree.SubElement(clefnode, "clef-octave-change")
            octchnode.text = str(oct_ch)

    def add_bar_style(self, multirest=None):
        bar_style = etree.SubElement(self.bar_attr, "measure-style")
        if multirest:
            multirestnode = etree.SubElement(bar_style, "multiple-rest")
            multirestnode.text = str(multirest)

    def new_system(self, force_break):
        etree.SubElement(self.current_bar, "print", {'new-system':force_break})

    def add_barline(self, bl_type, repeat=None):
        barnode = etree.SubElement(self.current_bar, "barline", location="right")
        barstyle = etree.SubElement(barnode, "bar-style")
        barstyle.text = bl_type
        if repeat:
            repeatnode = etree.SubElement(barnode, "repeat", direction=repeat)

    def add_backup(self, duration):
        if duration <= 0:
            return
        backupnode = etree.SubElement(self.current_bar, "backup")
        durnode = etree.SubElement(backupnode, "duration")
        durnode.text = str(duration)

    def add_voice(self, voice):
        voicenode = etree.SubElement(self.current_note, "voice")
        voicenode.text = str(voice)

    def add_staff(self, staff):
        staffnode = etree.SubElement(self.current_note, "staff")
        staffnode.text = str(staff)

    def add_staves(self, staves):
        index = get_tag_index(self.bar_attr, "time")
        stavesnode = etree.Element("staves")
        stavesnode.text = str(staves)
        self.bar_attr.insert(index + 1, stavesnode)

    def add_chord(self):
        etree.SubElement(self.current_note, "chord")

    def add_direction(self, pos="above"):
        self.direction = etree.SubElement(self.current_bar, "direction", placement=pos)

    def add_dynamic_mark(self, dyn):
        """Add specified dynamic mark."""
        direction = etree.SubElement(self.current_bar, "direction", placement='below')
        dirtypenode = etree.SubElement(direction, "direction-type")
        dyn_node = etree.SubElement(dirtypenode, "dynamics")
        dynexpr_node = etree.SubElement(dyn_node, dyn)

    def add_dynamic_wedge(self, wedge_type):
        """Add dynamic wedge/hairpin."""
        direction = etree.SubElement(self.current_bar, "direction", placement='below')
        dirtypenode = etree.SubElement(direction, "direction-type")
        dyn_node = etree.SubElement(dirtypenode, "wedge", type=wedge_type)

    def add_dynamic_text(self, text):
        """Add dynamic text."""
        direction = etree.SubElement(self.current_bar, "direction", placement='below')
        dirtypenode = etree.SubElement(direction, "direction-type")
        dyn_node = etree.SubElement(dirtypenode, "words")
        dyn_node.attrib['font-style'] = 'italic'
        dyn_node.text = text

    def add_dynamic_dashes(self, text):
        """Add dynamics dashes."""
        direction = etree.SubElement(self.current_bar, "direction", placement='below')
        dirtypenode = etree.SubElement(direction, "direction-type")
        dyn_node = etree.SubElement(dirtypenode, "dashes", type=text)

    def add_octave_shift(self, plac, octdir, size):
        """Add octave shift."""
        oct_dict = {"type": octdir, "size": str(size) }
        direction = etree.SubElement(self.current_bar, "direction", placement=plac)
        dirtypenode = etree.SubElement(direction, "direction-type")
        dyn_node = etree.SubElement(dirtypenode, "octave-shift", oct_dict)

    def add_dirwords(self, words):
        """Add words in direction, e. g. a tempo mark."""
        if self.current_bar.find('direction') is None:
            self.add_direction()
        dirtypenode = etree.SubElement(self.direction, "direction-type")
        wordsnode = etree.SubElement(dirtypenode, "words")
        wordsnode.text = words

    def add_mark(self, mark):
        """Add rehearsal mark in direction"""
        if self.current_bar.find('direction') is None:
            self.add_direction()
        dirtypenode = etree.SubElement(self.direction, "direction-type")
        rehearsalnode = etree.SubElement(dirtypenode, "rehearsal")
        rehearsalnode.text = mark

    def add_metron_dir(self, unit, beats, dots):
        dirtypenode = etree.SubElement(self.direction, "direction-type")
        metrnode = etree.SubElement(dirtypenode, "metronome")
        bunode = etree.SubElement(metrnode, "beat-unit")
        bunode.text = unit
        if dots:
            for d in range(dots):
                etree.SubElement(metrnode, "beat-unit-dot")
        pmnode = etree.SubElement(metrnode, "per-minute")
        pmnode.text = str(beats)

    def add_sound_dir(self, midi_tempo):
        # FIXME: remove the int conversion once LilyPond accepts decimal tempo
        soundnode = etree.SubElement(self.direction, "sound", tempo=str(int(midi_tempo)))

    def add_lyric(self, txt, syll, nr, ext=False):
        """ Add lyric element. """
        lyricnode = etree.SubElement(self.current_note, "lyric", number=str(nr))
        syllnode = etree.SubElement(lyricnode, "syllabic")
        syllnode.text = syll
        txtnode = etree.SubElement(lyricnode, "text")
        txtnode.text = txt
        if ext:
            etree.SubElement(lyricnode, "extend")


    ##
    # Create the XML document
    ##

    def musicxml(self, prettyprint=True):
        xml = MusicXML(self.tree)
        if prettyprint:
            xml.indent("  ")
        return xml


class MusicXML(object):
    """Represent a generated MusicXML tree."""
    def __init__(self, tree):
        self.tree = tree
        self.root = tree.getroot()

    def indent(self, indent="  "):
        """Add indent and linebreaks to the created XML tree."""
        import ly.etreeutil
        ly.etreeutil.indent(self.root, indent)

    def tostring(self, encoding='UTF-8'):
        """Output etree as a XML document."""
        return etree.tostring(self.root, encoding=encoding, method="xml")

    def write(self, file, encoding='UTF-8', doctype=True):
        """Write XML to a file (file obj or filename)."""
        def write(f):
            if doctype:
                f.write((xml_decl_txt + "\n").format(encoding=encoding).encode(encoding))
                f.write((doctype_txt + "\n").encode(encoding))
                self.tree.write(f, encoding=encoding, xml_declaration=False)
            else:
                self.tree.write(f, encoding=encoding, xml_declaration=True, method="xml")
        if hasattr(file, 'write'):
            write(file)
            # do not close if it already was a file object
        else:
            # it is not a file object
            with open(file, 'wb') as f:
                write(f)


def get_tag_index(node, tag):
    """Return the (first) index of tag in node.

    If tag is not found, -1 is returned.
    """
    for i, elem in enumerate(list(node)):
        if elem.tag == tag:
            return i
    return -1


xml_decl_txt = """<?xml version="1.0" encoding="{encoding}"?>"""

doctype_txt = """<!DOCTYPE score-partwise PUBLIC "-//Recordare//DTD MusicXML 2.0 Partwise//EN"
                                "http://www.musicxml.org/dtds/partwise.dtd">"""



# generated by LilyPond 2.24.0

version = "2.24.0"

interfaces = {
    "accidental-interface": [
        "alteration",
        "alteration-glyph-name-alist",
        "avoid-slur",
        "glyph-name",
        "hide-tied-accidental-after-break",
        "restore-first",
    ],
    "accidental-placement-interface": [
        "direction",
        "padding",
        "positioning-done",
        "right-padding",
        "script-priority",
    ],
    "accidental-suggestion-interface": [],
    "accidental-switch-interface": [
        "alteration-glyph-name-alist",
    ],
    "align-interface": [
        "align-dir",
        "axes",
        "padding",
        "positioning-done",
        "stacking-dir",
    ],
    "ambitus-interface": [
        "gap",
        "length-fraction",
        "maximum-gap",
        "thickness",
    ],
    "arpeggio-interface": [
        "arpeggio-direction",
        "dash-definition",
        "line-thickness",
        "positions",
        "protrusion",
        "script-priority",
        "thickness",
    ],
    "axis-group-interface": [
        "axes",
        "default-staff-staff-spacing",
        "nonstaff-nonstaff-spacing",
        "nonstaff-relatedstaff-spacing",
        "nonstaff-unrelatedstaff-spacing",
        "staff-affinity",
        "staff-staff-spacing",
    ],
    "balloon-interface": [
        "X-attachment",
        "Y-attachment",
        "annotation-balloon",
        "annotation-line",
        "padding",
        "text",
        "text-alignment-X",
        "text-alignment-Y",
        "thickness",
    ],
    "bar-line-interface": [
        "allow-span-bar",
        "bar-extent",
        "gap",
        "glyph",
        "glyph-left",
        "glyph-name",
        "glyph-right",
        "hair-thickness",
        "kern",
        "rounded",
        "segno-kern",
        "thick-thickness",
    ],
    "bar-number-interface": [],
    "bass-figure-alignment-interface": [],
    "bass-figure-interface": [
        "implicit",
    ],
    "beam-interface": [
        "X-positions",
        "auto-knee-gap",
        "beam-thickness",
        "beamed-stem-shorten",
        "beaming",
        "break-overshoot",
        "clip-edges",
        "collision-interfaces",
        "collision-voice-only",
        "concaveness",
        "damping",
        "details",
        "direction",
        "gap",
        "gap-count",
        "grow-direction",
        "inspect-quants",
        "knee",
        "length-fraction",
        "neutral-direction",
        "positions",
        "skip-quanting",
    ],
    "bend-after-interface": [
        "thickness",
    ],
    "bend-interface": [
        "bend-me",
        "details",
        "direction",
        "style",
    ],
    "bezier-curve-interface": [
        "show-control-points",
    ],
    "break-alignable-interface": [
        "break-align-symbols",
        "non-break-align-symbols",
    ],
    "break-aligned-interface": [
        "break-align-anchor",
        "break-align-anchor-alignment",
        "break-align-symbol",
        "space-alist",
    ],
    "break-alignment-interface": [
        "break-align-orders",
        "positioning-done",
    ],
    "breathing-sign-interface": [
        "direction",
        "thickness",
    ],
    "caesura-script-interface": [],
    "centered-bar-number-interface": [],
    "centered-bar-number-line-spanner-interface": [],
    "centered-spanner-interface": [
        "self-alignment-X",
        "spacing-pair",
    ],
    "chord-name-interface": [],
    "chord-square-interface": [
        "measure-division",
        "measure-division-chord-placement-alist",
        "measure-division-lines-alist",
    ],
    "clef-interface": [
        "full-size-change",
        "glyph",
        "glyph-name",
        "non-default",
    ],
    "clef-modifier-interface": [
        "clef-alignments",
    ],
    "cluster-beacon-interface": [
        "positions",
    ],
    "cluster-interface": [
        "padding",
        "style",
    ],
    "coda-mark-interface": [],
    "control-point-interface": [],
    "control-polygon-interface": [
        "extroversion",
        "filled",
    ],
    "custos-interface": [
        "neutral-direction",
        "neutral-position",
        "style",
    ],
    "dot-column-interface": [
        "chord-dots-limit",
        "direction",
        "padding",
        "positioning-done",
    ],
    "dots-interface": [
        "direction",
        "dot-count",
        "glyph-name",
        "style",
    ],
    "duration-line-interface": [
        "details",
    ],
    "dynamic-interface": [],
    "dynamic-line-spanner-interface": [
        "avoid-slur",
    ],
    "dynamic-text-interface": [
        "right-padding",
    ],
    "dynamic-text-spanner-interface": [
        "text",
    ],
    "enclosing-bracket-interface": [
        "bracket-flare",
        "dashed-edge",
        "edge-height",
        "padding",
        "shorten-pair",
        "thickness",
    ],
    "episema-interface": [],
    "figured-bass-continuation-interface": [
        "padding",
        "thickness",
    ],
    "finger-glide-interface": [
        "dash-fraction",
        "dash-period",
        "details",
        "style",
        "zigzag-length",
        "zigzag-width",
    ],
    "finger-interface": [],
    "fingering-column-interface": [
        "padding",
        "positioning-done",
        "snap-radius",
    ],
    "flag-interface": [
        "glyph-name",
        "stroke-style",
        "style",
    ],
    "font-interface": [
        "font-encoding",
        "font-family",
        "font-features",
        "font-name",
        "font-series",
        "font-shape",
        "font-size",
    ],
    "footnote-interface": [
        "automatically-numbered",
        "footnote",
        "footnote-text",
    ],
    "fret-diagram-interface": [
        "align-dir",
        "dot-placement-list",
        "fret-diagram-details",
        "size",
        "thickness",
    ],
    "glissando-interface": [],
    "grace-spacing-interface": [
        "common-shortest-duration",
    ],
    "gregorian-ligature-interface": [],
    "grid-chord-name-interface": [],
    "grid-line-interface": [
        "thickness",
    ],
    "grid-point-interface": [],
    "grob-interface": [
        "X-extent",
        "X-offset",
        "Y-extent",
        "Y-offset",
        "after-line-breaking",
        "avoid-slur",
        "before-line-breaking",
        "color",
        "cross-staff",
        "extra-offset",
        "footnote-music",
        "forced-spacing",
        "horizontal-skylines",
        "id",
        "layer",
        "minimum-X-extent",
        "minimum-Y-extent",
        "output-attributes",
        "parenthesis-friends",
        "parenthesis-id",
        "parenthesized",
        "rotation",
        "show-horizontal-skylines",
        "show-vertical-skylines",
        "skyline-horizontal-padding",
        "springs-and-rods",
        "stencil",
        "transparent",
        "vertical-skylines",
        "whiteout",
        "whiteout-style",
    ],
    "hairpin-interface": [
        "bound-padding",
        "broken-bound-padding",
        "circled-tip",
        "endpoint-alignments",
        "grow-direction",
        "height",
        "shorten-pair",
    ],
    "hara-kiri-group-spanner-interface": [
        "remove-empty",
        "remove-first",
        "remove-layer",
    ],
    "horizontal-bracket-interface": [
        "bracket-flare",
        "connect-to-neighbor",
        "dashed-edge",
        "edge-height",
        "shorten-pair",
    ],
    "horizontal-bracket-text-interface": [],
    "horizontal-line-spanner-interface": [],
    "inline-accidental-interface": [],
    "instrument-specific-markup-interface": [
        "fret-diagram-details",
        "graphical",
        "harp-pedal-details",
        "size",
        "thickness",
        "woodwind-diagram-details",
    ],
    "item-interface": [
        "break-visibility",
        "extra-spacing-height",
        "extra-spacing-width",
        "non-musical",
    ],
    "jump-script-interface": [],
    "key-cancellation-interface": [],
    "key-signature-interface": [
        "alteration-alist",
        "alteration-glyph-name-alist",
        "flat-positions",
        "non-default",
        "padding",
        "padding-pairs",
        "sharp-positions",
    ],
    "kievan-ligature-interface": [
        "padding",
    ],
    "ledger-line-spanner-interface": [
        "gap",
        "length-fraction",
        "minimum-length-fraction",
    ],
    "ledgered-interface": [
        "no-ledgers",
    ],
    "ligature-bracket-interface": [
        "height",
        "thickness",
        "width",
    ],
    "ligature-head-interface": [],
    "ligature-interface": [],
    "line-interface": [
        "arrow-length",
        "arrow-width",
        "dash-fraction",
        "dash-period",
        "style",
        "thickness",
        "zigzag-length",
        "zigzag-width",
    ],
    "line-spanner-interface": [
        "bound-details",
        "extra-dy",
        "gap",
        "left-bound-info",
        "right-bound-info",
        "thickness",
        "to-barline",
    ],
    "lyric-extender-interface": [
        "left-padding",
        "next",
        "right-padding",
        "thickness",
    ],
    "lyric-hyphen-interface": [
        "dash-period",
        "height",
        "length",
        "minimum-distance",
        "minimum-length",
        "padding",
        "thickness",
    ],
    "lyric-interface": [],
    "lyric-repeat-count-interface": [],
    "lyric-space-interface": [],
    "lyric-syllable-interface": [],
    "mark-interface": [],
    "measure-counter-interface": [
        "count-from",
        "left-number-text",
        "number-range-separator",
        "right-number-text",
    ],
    "measure-grouping-interface": [
        "height",
        "style",
        "thickness",
    ],
    "measure-spanner-interface": [
        "bracket-flare",
        "bracket-visibility",
        "connect-to-neighbor",
        "direction",
        "edge-height",
        "padding",
        "shorten-pair",
        "spacing-pair",
        "staff-padding",
        "thickness",
    ],
    "melody-spanner-interface": [
        "neutral-direction",
    ],
    "mensural-ligature-interface": [
        "thickness",
    ],
    "metronome-mark-interface": [],
    "multi-measure-interface": [
        "bound-padding",
    ],
    "multi-measure-rest-interface": [
        "bound-padding",
        "expand-limit",
        "hair-thickness",
        "max-symbol-separation",
        "measure-count",
        "minimum-length",
        "round-up-exceptions",
        "round-up-to-longer-rest",
        "spacing-pair",
        "thick-thickness",
        "usable-duration-logs",
    ],
    "multi-measure-rest-number-interface": [],
    "musical-paper-column-interface": [
        "shortest-playing-duration",
        "shortest-starter-duration",
    ],
    "non-musical-paper-column-interface": [
        "between-cols",
        "full-measure-extra-space",
        "line-break-penalty",
        "line-break-permission",
        "line-break-system-details",
        "page-break-penalty",
        "page-break-permission",
        "page-turn-penalty",
        "page-turn-permission",
    ],
    "note-collision-interface": [
        "fa-merge-direction",
        "merge-differently-dotted",
        "merge-differently-headed",
        "note-collision-threshold",
        "positioning-done",
        "prefer-dotted-right",
    ],
    "note-column-interface": [
        "force-hshift",
        "glissando-skip",
        "horizontal-shift",
        "ignore-collision",
        "main-extent",
    ],
    "note-head-interface": [
        "duration-log",
        "glyph-name",
        "ignore-ambitus",
        "ledger-positions",
        "note-names",
        "stem-attachment",
        "style",
    ],
    "note-name-interface": [],
    "note-spacing-interface": [
        "knee-spacing-correction",
        "same-direction-correction",
        "space-to-barline",
        "stem-spacing-correction",
    ],
    "number-interface": [
        "number-type",
    ],
    "ottava-bracket-interface": [
        "bracket-flare",
        "dashed-edge",
        "edge-height",
        "minimum-length",
        "shorten-pair",
    ],
    "outside-staff-axis-group-interface": [
        "outside-staff-placement-directive",
    ],
    "outside-staff-interface": [
        "outside-staff-horizontal-padding",
        "outside-staff-padding",
        "outside-staff-priority",
    ],
    "paper-column-interface": [
        "labels",
        "rhythmic-location",
        "used",
        "when",
    ],
    "parentheses-interface": [
        "padding",
        "stencils",
    ],
    "percent-repeat-interface": [
        "dot-negative-kern",
        "slash-negative-kern",
        "slope",
        "thickness",
    ],
    "piano-pedal-bracket-interface": [
        "bound-padding",
        "bracket-flare",
        "dashed-edge",
        "edge-height",
        "shorten-pair",
    ],
    "piano-pedal-interface": [],
    "piano-pedal-script-interface": [],
    "pitched-trill-interface": [],
    "pure-from-neighbor-interface": [],
    "rehearsal-mark-interface": [],
    "rest-collision-interface": [
        "minimum-distance",
        "positioning-done",
    ],
    "rest-interface": [
        "direction",
        "minimum-distance",
        "style",
        "voiced-position",
    ],
    "rhythmic-grob-interface": [],
    "rhythmic-head-interface": [
        "duration-log",
        "glissando-skip",
    ],
    "script-column-interface": [],
    "script-interface": [
        "avoid-slur",
        "positioning-done",
        "script-priority",
        "side-relative-direction",
        "slur-padding",
        "toward-stem-shift",
        "toward-stem-shift-in-column",
    ],
    "section-label-interface": [],
    "segno-mark-interface": [],
    "self-alignment-interface": [
        "X-align-on-main-noteheads",
        "parent-alignment-X",
        "parent-alignment-Y",
        "self-alignment-X",
        "self-alignment-Y",
    ],
    "semi-tie-column-interface": [
        "head-direction",
        "positioning-done",
        "tie-configuration",
    ],
    "semi-tie-interface": [
        "control-points",
        "details",
        "direction",
        "head-direction",
        "line-thickness",
        "thickness",
    ],
    "separation-item-interface": [
        "X-extent",
        "horizontal-skylines",
        "padding",
        "skyline-vertical-padding",
    ],
    "side-position-interface": [
        "add-stem-support",
        "direction",
        "horizon-padding",
        "minimum-space",
        "padding",
        "side-axis",
        "slur-padding",
        "staff-padding",
        "use-skylines",
    ],
    "signum-repetitionis-interface": [
        "allow-span-bar",
        "bar-extent",
        "gap",
        "glyph",
        "glyph-name",
        "hair-thickness",
        "kern",
        "rounded",
        "segno-kern",
        "thick-thickness",
    ],
    "slur-interface": [
        "avoid-slur",
        "control-points",
        "dash-definition",
        "details",
        "direction",
        "eccentricity",
        "height-limit",
        "inspect-quants",
        "line-thickness",
        "positions",
        "ratio",
        "thickness",
    ],
    "spaceable-grob-interface": [
        "allow-loose-spacing",
        "keep-inside-line",
        "measure-length",
    ],
    "spacing-interface": [],
    "spacing-options-interface": [
        "shortest-duration-space",
        "spacing-increment",
    ],
    "spacing-spanner-interface": [
        "average-spacing-wishes",
        "base-shortest-duration",
        "common-shortest-duration",
        "packed-spacing",
        "shortest-duration-space",
        "spacing-increment",
        "strict-grace-spacing",
        "strict-note-spacing",
        "uniform-stretching",
    ],
    "span-bar-interface": [
        "glyph-name",
    ],
    "spanner-interface": [
        "minimum-length",
        "minimum-length-after-break",
        "normalized-endpoints",
        "spanner-id",
        "to-barline",
    ],
    "staff-grouper-interface": [
        "staff-staff-spacing",
        "staffgroup-staff-spacing",
    ],
    "staff-highlight-interface": [
        "bound-prefatory-paddings",
        "shorten-pair",
    ],
    "staff-spacing-interface": [
        "stem-spacing-correction",
    ],
    "staff-symbol-interface": [
        "break-align-symbols",
        "ledger-extra",
        "ledger-line-thickness",
        "ledger-positions",
        "ledger-positions-function",
        "line-count",
        "line-positions",
        "staff-space",
        "thickness",
        "widened-extent",
        "width",
    ],
    "staff-symbol-referencer-interface": [
        "staff-position",
    ],
    "stanza-number-interface": [],
    "stem-interface": [
        "avoid-note-head",
        "beaming",
        "beamlet-default-length",
        "beamlet-max-length-proportion",
        "default-direction",
        "details",
        "direction",
        "double-stem-separation",
        "duration-log",
        "french-beaming",
        "length",
        "length-fraction",
        "max-beam-connect",
        "neutral-direction",
        "no-stem-extend",
        "note-collision-threshold",
        "positioning-done",
        "stem-begin-position",
        "stemlet-length",
        "thickness",
    ],
    "stem-tremolo-interface": [
        "beam-thickness",
        "beam-width",
        "direction",
        "flag-count",
        "length-fraction",
        "shape",
        "slope",
    ],
    "sticky-grob-interface": [],
    "string-number-interface": [],
    "stroke-finger-interface": [
        "digit-names",
    ],
    "system-interface": [
        "labels",
        "page-number",
        "rank-on-page",
    ],
    "system-start-delimiter-interface": [
        "collapse-height",
        "style",
        "thickness",
    ],
    "system-start-text-interface": [
        "long-text",
        "self-alignment-X",
        "self-alignment-Y",
        "text",
    ],
    "tab-note-head-interface": [
        "details",
    ],
    "text-interface": [
        "baseline-skip",
        "flag-style",
        "replacement-alist",
        "text",
        "text-direction",
        "word-space",
    ],
    "text-mark-interface": [],
    "text-script-interface": [
        "avoid-slur",
        "script-priority",
    ],
    "tie-column-interface": [
        "positioning-done",
        "tie-configuration",
    ],
    "tie-interface": [
        "avoid-slur",
        "control-points",
        "dash-definition",
        "details",
        "direction",
        "head-direction",
        "line-thickness",
        "neutral-direction",
        "staff-position",
        "thickness",
    ],
    "time-signature-interface": [
        "fraction",
        "style",
    ],
    "trill-pitch-accidental-interface": [],
    "trill-spanner-interface": [],
    "tuplet-bracket-interface": [
        "X-positions",
        "avoid-scripts",
        "bracket-flare",
        "bracket-visibility",
        "break-overshoot",
        "connect-to-neighbor",
        "dashed-edge",
        "direction",
        "edge-height",
        "edge-text",
        "full-length-padding",
        "full-length-to-extent",
        "gap",
        "padding",
        "positions",
        "shorten-pair",
        "staff-padding",
        "thickness",
        "tuplet-slur",
        "visible-over-note-heads",
    ],
    "tuplet-number-interface": [
        "avoid-slur",
        "direction",
        "knee-to-beam",
    ],
    "unbreakable-spanner-interface": [
        "breakable",
    ],
    "vaticana-ligature-interface": [
        "glyph-name",
        "thickness",
    ],
    "volta-bracket-interface": [
        "dashed-edge",
        "height",
        "shorten-pair",
        "thickness",
    ],
    "volta-interface": [],
}


grobs = {
    "Accidental": [
        "accidental-interface",
        "accidental-switch-interface",
        "font-interface",
        "grob-interface",
        "inline-accidental-interface",
        "item-interface",
    ],
    "AccidentalCautionary": [
        "accidental-interface",
        "accidental-switch-interface",
        "font-interface",
        "grob-interface",
        "inline-accidental-interface",
        "item-interface",
    ],
    "AccidentalPlacement": [
        "accidental-placement-interface",
        "grob-interface",
        "item-interface",
    ],
    "AccidentalSuggestion": [
        "accidental-interface",
        "accidental-suggestion-interface",
        "accidental-switch-interface",
        "font-interface",
        "grob-interface",
        "item-interface",
        "outside-staff-interface",
        "script-interface",
        "self-alignment-interface",
        "side-position-interface",
    ],
    "Ambitus": [
        "ambitus-interface",
        "axis-group-interface",
        "break-aligned-interface",
        "grob-interface",
        "item-interface",
    ],
    "AmbitusAccidental": [
        "accidental-interface",
        "accidental-switch-interface",
        "break-aligned-interface",
        "font-interface",
        "grob-interface",
        "item-interface",
    ],
    "AmbitusLine": [
        "ambitus-interface",
        "font-interface",
        "grob-interface",
        "item-interface",
    ],
    "AmbitusNoteHead": [
        "ambitus-interface",
        "font-interface",
        "grob-interface",
        "item-interface",
        "ledgered-interface",
        "note-head-interface",
        "rhythmic-head-interface",
        "staff-symbol-referencer-interface",
    ],
    "Arpeggio": [
        "arpeggio-interface",
        "font-interface",
        "grob-interface",
        "item-interface",
        "side-position-interface",
        "staff-symbol-referencer-interface",
    ],
    "BalloonText": [
        "accidental-switch-interface",
        "balloon-interface",
        "font-interface",
        "grob-interface",
        "item-interface",
        "spanner-interface",
        "sticky-grob-interface",
        "text-interface",
    ],
    "BarLine": [
        "bar-line-interface",
        "break-aligned-interface",
        "font-interface",
        "grob-interface",
        "item-interface",
        "pure-from-neighbor-interface",
    ],
    "BarNumber": [
        "bar-number-interface",
        "break-alignable-interface",
        "font-interface",
        "grob-interface",
        "item-interface",
        "outside-staff-interface",
        "self-alignment-interface",
        "side-position-interface",
        "text-interface",
    ],
    "BassFigure": [
        "accidental-switch-interface",
        "bass-figure-interface",
        "font-interface",
        "grob-interface",
        "item-interface",
        "rhythmic-grob-interface",
        "text-interface",
    ],
    "BassFigureAlignment": [
        "align-interface",
        "axis-group-interface",
        "bass-figure-alignment-interface",
        "grob-interface",
        "spanner-interface",
    ],
    "BassFigureAlignmentPositioning": [
        "axis-group-interface",
        "grob-interface",
        "outside-staff-interface",
        "side-position-interface",
        "spanner-interface",
    ],
    "BassFigureBracket": [
        "enclosing-bracket-interface",
        "grob-interface",
        "item-interface",
    ],
    "BassFigureContinuation": [
        "figured-bass-continuation-interface",
        "grob-interface",
        "spanner-interface",
    ],
    "BassFigureLine": [
        "axis-group-interface",
        "grob-interface",
        "outside-staff-axis-group-interface",
        "spanner-interface",
    ],
    "Beam": [
        "beam-interface",
        "grob-interface",
        "spanner-interface",
        "staff-symbol-referencer-interface",
        "unbreakable-spanner-interface",
    ],
    "BendAfter": [
        "bend-after-interface",
        "grob-interface",
        "spanner-interface",
    ],
    "BendSpanner": [
        "bend-interface",
        "font-interface",
        "grob-interface",
        "line-spanner-interface",
        "outside-staff-interface",
        "spanner-interface",
        "text-interface",
        "text-script-interface",
    ],
    "BreakAlignGroup": [
        "axis-group-interface",
        "break-aligned-interface",
        "grob-interface",
        "item-interface",
    ],
    "BreakAlignment": [
        "axis-group-interface",
        "break-alignment-interface",
        "grob-interface",
        "item-interface",
    ],
    "BreathingSign": [
        "break-aligned-interface",
        "breathing-sign-interface",
        "font-interface",
        "grob-interface",
        "item-interface",
        "outside-staff-interface",
        "text-interface",
    ],
    "CaesuraScript": [
        "caesura-script-interface",
        "font-interface",
        "grob-interface",
        "item-interface",
        "outside-staff-interface",
        "script-interface",
        "self-alignment-interface",
        "side-position-interface",
    ],
    "CenteredBarNumber": [
        "bar-number-interface",
        "centered-bar-number-interface",
        "centered-spanner-interface",
        "font-interface",
        "grob-interface",
        "spanner-interface",
        "text-interface",
    ],
    "CenteredBarNumberLineSpanner": [
        "axis-group-interface",
        "bar-number-interface",
        "centered-bar-number-line-spanner-interface",
        "grob-interface",
        "outside-staff-interface",
        "side-position-interface",
        "spanner-interface",
    ],
    "ChordName": [
        "accidental-switch-interface",
        "chord-name-interface",
        "font-interface",
        "grob-interface",
        "item-interface",
        "outside-staff-interface",
        "rhythmic-grob-interface",
        "text-interface",
    ],
    "ChordSquare": [
        "chord-square-interface",
        "grob-interface",
        "line-interface",
        "spanner-interface",
    ],
    "Clef": [
        "break-aligned-interface",
        "clef-interface",
        "font-interface",
        "grob-interface",
        "item-interface",
        "pure-from-neighbor-interface",
        "staff-symbol-referencer-interface",
    ],
    "ClefModifier": [
        "clef-modifier-interface",
        "font-interface",
        "grob-interface",
        "item-interface",
        "outside-staff-interface",
        "self-alignment-interface",
        "side-position-interface",
        "text-interface",
    ],
    "ClusterSpanner": [
        "cluster-interface",
        "grob-interface",
        "spanner-interface",
    ],
    "ClusterSpannerBeacon": [
        "cluster-beacon-interface",
        "grob-interface",
        "item-interface",
        "rhythmic-grob-interface",
    ],
    "CodaMark": [
        "break-alignable-interface",
        "coda-mark-interface",
        "font-interface",
        "grob-interface",
        "item-interface",
        "mark-interface",
        "outside-staff-interface",
        "self-alignment-interface",
        "side-position-interface",
        "text-interface",
    ],
    "CombineTextScript": [
        "accidental-switch-interface",
        "font-interface",
        "grob-interface",
        "item-interface",
        "outside-staff-interface",
        "self-alignment-interface",
        "side-position-interface",
        "text-interface",
        "text-script-interface",
    ],
    "ControlPoint": [
        "control-point-interface",
        "grob-interface",
        "item-interface",
        "spanner-interface",
        "sticky-grob-interface",
        "text-interface",
    ],
    "ControlPolygon": [
        "control-polygon-interface",
        "grob-interface",
        "item-interface",
        "spanner-interface",
        "sticky-grob-interface",
        "text-interface",
    ],
    "CueClef": [
        "break-aligned-interface",
        "clef-interface",
        "font-interface",
        "grob-interface",
        "item-interface",
        "pure-from-neighbor-interface",
        "staff-symbol-referencer-interface",
    ],
    "CueEndClef": [
        "break-aligned-interface",
        "clef-interface",
        "font-interface",
        "grob-interface",
        "item-interface",
        "pure-from-neighbor-interface",
        "staff-symbol-referencer-interface",
    ],
    "Custos": [
        "break-aligned-interface",
        "custos-interface",
        "font-interface",
        "grob-interface",
        "item-interface",
        "staff-symbol-referencer-interface",
    ],
    "Divisio": [
        "break-aligned-interface",
        "breathing-sign-interface",
        "font-interface",
        "grob-interface",
        "item-interface",
        "outside-staff-interface",
        "text-interface",
    ],
    "DotColumn": [
        "axis-group-interface",
        "dot-column-interface",
        "grob-interface",
        "item-interface",
    ],
    "Dots": [
        "dots-interface",
        "font-interface",
        "grob-interface",
        "item-interface",
        "staff-symbol-referencer-interface",
    ],
    "DoublePercentRepeat": [
        "break-aligned-interface",
        "font-interface",
        "grob-interface",
        "item-interface",
        "percent-repeat-interface",
    ],
    "DoublePercentRepeatCounter": [
        "font-interface",
        "grob-interface",
        "item-interface",
        "outside-staff-interface",
        "self-alignment-interface",
        "side-position-interface",
        "text-interface",
    ],
    "DoubleRepeatSlash": [
        "font-interface",
        "grob-interface",
        "item-interface",
        "outside-staff-interface",
        "percent-repeat-interface",
        "rhythmic-grob-interface",
    ],
    "DurationLine": [
        "duration-line-interface",
        "font-interface",
        "grob-interface",
        "horizontal-line-spanner-interface",
        "line-interface",
        "line-spanner-interface",
        "spanner-interface",
        "unbreakable-spanner-interface",
    ],
    "DynamicLineSpanner": [
        "axis-group-interface",
        "dynamic-interface",
        "dynamic-line-spanner-interface",
        "grob-interface",
        "outside-staff-interface",
        "side-position-interface",
        "spanner-interface",
    ],
    "DynamicText": [
        "dynamic-interface",
        "dynamic-text-interface",
        "font-interface",
        "grob-interface",
        "item-interface",
        "outside-staff-interface",
        "script-interface",
        "self-alignment-interface",
        "text-interface",
    ],
    "DynamicTextSpanner": [
        "dynamic-interface",
        "dynamic-text-spanner-interface",
        "font-interface",
        "grob-interface",
        "horizontal-line-spanner-interface",
        "line-interface",
        "line-spanner-interface",
        "spanner-interface",
        "text-interface",
    ],
    "Episema": [
        "episema-interface",
        "font-interface",
        "grob-interface",
        "horizontal-line-spanner-interface",
        "line-interface",
        "line-spanner-interface",
        "side-position-interface",
        "spanner-interface",
    ],
    "FingerGlideSpanner": [
        "finger-glide-interface",
        "grob-interface",
        "line-spanner-interface",
        "spanner-interface",
    ],
    "Fingering": [
        "finger-interface",
        "font-interface",
        "grob-interface",
        "item-interface",
        "outside-staff-interface",
        "self-alignment-interface",
        "side-position-interface",
        "text-interface",
        "text-script-interface",
    ],
    "FingeringColumn": [
        "fingering-column-interface",
        "grob-interface",
        "item-interface",
    ],
    "Flag": [
        "flag-interface",
        "font-interface",
        "grob-interface",
        "item-interface",
    ],
    "Footnote": [
        "balloon-interface",
        "font-interface",
        "footnote-interface",
        "grob-interface",
        "item-interface",
        "spanner-interface",
        "sticky-grob-interface",
        "text-interface",
    ],
    "FretBoard": [
        "chord-name-interface",
        "font-interface",
        "fret-diagram-interface",
        "grob-interface",
        "item-interface",
        "outside-staff-interface",
        "rhythmic-grob-interface",
    ],
    "Glissando": [
        "glissando-interface",
        "grob-interface",
        "line-interface",
        "line-spanner-interface",
        "spanner-interface",
        "unbreakable-spanner-interface",
    ],
    "GraceSpacing": [
        "grace-spacing-interface",
        "grob-interface",
        "spacing-options-interface",
        "spanner-interface",
    ],
    "GridChordName": [
        "accidental-switch-interface",
        "font-interface",
        "grid-chord-name-interface",
        "grob-interface",
        "spanner-interface",
        "text-interface",
    ],
    "GridLine": [
        "grid-line-interface",
        "grob-interface",
        "item-interface",
        "self-alignment-interface",
    ],
    "GridPoint": [
        "grid-point-interface",
        "grob-interface",
        "item-interface",
    ],
    "Hairpin": [
        "dynamic-interface",
        "grob-interface",
        "hairpin-interface",
        "line-interface",
        "outside-staff-interface",
        "self-alignment-interface",
        "spanner-interface",
    ],
    "HorizontalBracket": [
        "grob-interface",
        "horizontal-bracket-interface",
        "line-interface",
        "outside-staff-interface",
        "side-position-interface",
        "spanner-interface",
    ],
    "HorizontalBracketText": [
        "accidental-switch-interface",
        "font-interface",
        "grob-interface",
        "horizontal-bracket-text-interface",
        "outside-staff-interface",
        "self-alignment-interface",
        "side-position-interface",
        "spanner-interface",
        "text-interface",
    ],
    "InstrumentName": [
        "accidental-switch-interface",
        "font-interface",
        "grob-interface",
        "self-alignment-interface",
        "side-position-interface",
        "spanner-interface",
        "system-start-text-interface",
        "text-interface",
    ],
    "InstrumentSwitch": [
        "accidental-switch-interface",
        "font-interface",
        "grob-interface",
        "item-interface",
        "outside-staff-interface",
        "self-alignment-interface",
        "side-position-interface",
        "text-interface",
    ],
    "JumpScript": [
        "break-alignable-interface",
        "font-interface",
        "grob-interface",
        "item-interface",
        "jump-script-interface",
        "outside-staff-interface",
        "self-alignment-interface",
        "side-position-interface",
        "text-interface",
    ],
    "KeyCancellation": [
        "accidental-switch-interface",
        "break-aligned-interface",
        "font-interface",
        "grob-interface",
        "item-interface",
        "key-cancellation-interface",
        "key-signature-interface",
        "pure-from-neighbor-interface",
        "staff-symbol-referencer-interface",
    ],
    "KeySignature": [
        "accidental-switch-interface",
        "break-aligned-interface",
        "font-interface",
        "grob-interface",
        "item-interface",
        "key-signature-interface",
        "pure-from-neighbor-interface",
        "staff-symbol-referencer-interface",
    ],
    "KievanLigature": [
        "font-interface",
        "grob-interface",
        "kievan-ligature-interface",
        "spanner-interface",
    ],
    "LaissezVibrerTie": [
        "bezier-curve-interface",
        "grob-interface",
        "item-interface",
        "semi-tie-interface",
        "tie-interface",
    ],
    "LaissezVibrerTieColumn": [
        "grob-interface",
        "item-interface",
        "semi-tie-column-interface",
    ],
    "LedgerLineSpanner": [
        "grob-interface",
        "ledger-line-spanner-interface",
        "spanner-interface",
    ],
    "LeftEdge": [
        "break-aligned-interface",
        "grob-interface",
        "item-interface",
    ],
    "LigatureBracket": [
        "grob-interface",
        "line-interface",
        "spanner-interface",
        "tuplet-bracket-interface",
    ],
    "LyricExtender": [
        "grob-interface",
        "lyric-extender-interface",
        "lyric-interface",
        "spanner-interface",
    ],
    "LyricHyphen": [
        "font-interface",
        "grob-interface",
        "lyric-hyphen-interface",
        "lyric-interface",
        "spanner-interface",
    ],
    "LyricRepeatCount": [
        "break-alignable-interface",
        "font-interface",
        "grob-interface",
        "item-interface",
        "lyric-interface",
        "lyric-repeat-count-interface",
        "self-alignment-interface",
        "text-interface",
    ],
    "LyricSpace": [
        "grob-interface",
        "lyric-hyphen-interface",
        "lyric-space-interface",
        "spanner-interface",
    ],
    "LyricText": [
        "font-interface",
        "grob-interface",
        "item-interface",
        "lyric-syllable-interface",
        "rhythmic-grob-interface",
        "self-alignment-interface",
        "text-interface",
    ],
    "MeasureCounter": [
        "centered-spanner-interface",
        "font-interface",
        "grob-interface",
        "measure-counter-interface",
        "outside-staff-interface",
        "self-alignment-interface",
        "side-position-interface",
        "spanner-interface",
        "text-interface",
    ],
    "MeasureGrouping": [
        "grob-interface",
        "measure-grouping-interface",
        "outside-staff-interface",
        "side-position-interface",
        "spanner-interface",
    ],
    "MeasureSpanner": [
        "accidental-switch-interface",
        "font-interface",
        "grob-interface",
        "line-interface",
        "measure-spanner-interface",
        "outside-staff-interface",
        "self-alignment-interface",
        "side-position-interface",
        "spanner-interface",
        "text-interface",
    ],
    "MelodyItem": [
        "grob-interface",
        "item-interface",
        "melody-spanner-interface",
    ],
    "MensuralLigature": [
        "font-interface",
        "grob-interface",
        "mensural-ligature-interface",
        "spanner-interface",
    ],
    "MetronomeMark": [
        "break-alignable-interface",
        "font-interface",
        "grob-interface",
        "item-interface",
        "metronome-mark-interface",
        "outside-staff-interface",
        "self-alignment-interface",
        "side-position-interface",
        "text-interface",
    ],
    "MultiMeasureRest": [
        "font-interface",
        "grob-interface",
        "multi-measure-interface",
        "multi-measure-rest-interface",
        "outside-staff-interface",
        "rest-interface",
        "spanner-interface",
        "staff-symbol-referencer-interface",
    ],
    "MultiMeasureRestNumber": [
        "font-interface",
        "grob-interface",
        "multi-measure-interface",
        "multi-measure-rest-number-interface",
        "outside-staff-interface",
        "self-alignment-interface",
        "side-position-interface",
        "spanner-interface",
        "text-interface",
    ],
    "MultiMeasureRestScript": [
        "font-interface",
        "grob-interface",
        "multi-measure-interface",
        "outside-staff-interface",
        "script-interface",
        "self-alignment-interface",
        "side-position-interface",
        "spanner-interface",
    ],
    "MultiMeasureRestText": [
        "font-interface",
        "grob-interface",
        "multi-measure-interface",
        "outside-staff-interface",
        "self-alignment-interface",
        "side-position-interface",
        "spanner-interface",
        "text-interface",
    ],
    "NonMusicalPaperColumn": [
        "axis-group-interface",
        "font-interface",
        "grob-interface",
        "item-interface",
        "non-musical-paper-column-interface",
        "paper-column-interface",
        "separation-item-interface",
        "spaceable-grob-interface",
    ],
    "NoteCollision": [
        "axis-group-interface",
        "grob-interface",
        "item-interface",
        "note-collision-interface",
    ],
    "NoteColumn": [
        "axis-group-interface",
        "bend-interface",
        "grob-interface",
        "item-interface",
        "note-column-interface",
        "separation-item-interface",
    ],
    "NoteHead": [
        "bend-interface",
        "font-interface",
        "gregorian-ligature-interface",
        "grob-interface",
        "item-interface",
        "ledgered-interface",
        "ligature-head-interface",
        "mensural-ligature-interface",
        "note-head-interface",
        "rhythmic-grob-interface",
        "rhythmic-head-interface",
        "staff-symbol-referencer-interface",
        "vaticana-ligature-interface",
    ],
    "NoteName": [
        "accidental-switch-interface",
        "font-interface",
        "grob-interface",
        "item-interface",
        "note-name-interface",
        "text-interface",
    ],
    "NoteSpacing": [
        "grob-interface",
        "item-interface",
        "note-spacing-interface",
        "spacing-interface",
    ],
    "OttavaBracket": [
        "font-interface",
        "grob-interface",
        "horizontal-bracket-interface",
        "line-interface",
        "ottava-bracket-interface",
        "outside-staff-interface",
        "side-position-interface",
        "spanner-interface",
        "text-interface",
    ],
    "PaperColumn": [
        "axis-group-interface",
        "font-interface",
        "grob-interface",
        "item-interface",
        "musical-paper-column-interface",
        "paper-column-interface",
        "separation-item-interface",
        "spaceable-grob-interface",
    ],
    "Parentheses": [
        "font-interface",
        "grob-interface",
        "item-interface",
        "parentheses-interface",
        "spanner-interface",
        "sticky-grob-interface",
    ],
    "PercentRepeat": [
        "centered-spanner-interface",
        "font-interface",
        "grob-interface",
        "multi-measure-rest-interface",
        "percent-repeat-interface",
        "spanner-interface",
    ],
    "PercentRepeatCounter": [
        "font-interface",
        "grob-interface",
        "outside-staff-interface",
        "self-alignment-interface",
        "side-position-interface",
        "spanner-interface",
        "text-interface",
    ],
    "PhrasingSlur": [
        "bezier-curve-interface",
        "grob-interface",
        "outside-staff-interface",
        "slur-interface",
        "spanner-interface",
    ],
    "PianoPedalBracket": [
        "grob-interface",
        "line-interface",
        "piano-pedal-bracket-interface",
        "piano-pedal-interface",
        "spanner-interface",
    ],
    "RehearsalMark": [
        "accidental-switch-interface",
        "break-alignable-interface",
        "font-interface",
        "grob-interface",
        "item-interface",
        "mark-interface",
        "outside-staff-interface",
        "rehearsal-mark-interface",
        "self-alignment-interface",
        "side-position-interface",
        "text-interface",
    ],
    "RepeatSlash": [
        "grob-interface",
        "item-interface",
        "percent-repeat-interface",
        "rhythmic-grob-interface",
    ],
    "RepeatTie": [
        "bezier-curve-interface",
        "grob-interface",
        "item-interface",
        "semi-tie-interface",
        "tie-interface",
    ],
    "RepeatTieColumn": [
        "grob-interface",
        "item-interface",
        "semi-tie-column-interface",
    ],
    "Rest": [
        "font-interface",
        "grob-interface",
        "item-interface",
        "rest-interface",
        "rhythmic-grob-interface",
        "rhythmic-head-interface",
        "staff-symbol-referencer-interface",
    ],
    "RestCollision": [
        "grob-interface",
        "item-interface",
        "rest-collision-interface",
    ],
    "Script": [
        "font-interface",
        "grob-interface",
        "item-interface",
        "outside-staff-interface",
        "script-interface",
        "self-alignment-interface",
        "side-position-interface",
    ],
    "ScriptColumn": [
        "grob-interface",
        "item-interface",
        "script-column-interface",
    ],
    "ScriptRow": [
        "grob-interface",
        "item-interface",
        "script-column-interface",
    ],
    "SectionLabel": [
        "break-alignable-interface",
        "font-interface",
        "grob-interface",
        "item-interface",
        "outside-staff-interface",
        "section-label-interface",
        "self-alignment-interface",
        "side-position-interface",
        "text-interface",
    ],
    "SegnoMark": [
        "break-alignable-interface",
        "font-interface",
        "grob-interface",
        "item-interface",
        "mark-interface",
        "outside-staff-interface",
        "segno-mark-interface",
        "self-alignment-interface",
        "side-position-interface",
        "text-interface",
    ],
    "SignumRepetitionis": [
        "break-aligned-interface",
        "font-interface",
        "grob-interface",
        "item-interface",
        "pure-from-neighbor-interface",
        "signum-repetitionis-interface",
    ],
    "Slur": [
        "bezier-curve-interface",
        "grob-interface",
        "outside-staff-interface",
        "slur-interface",
        "spanner-interface",
    ],
    "SostenutoPedal": [
        "font-interface",
        "grob-interface",
        "item-interface",
        "piano-pedal-script-interface",
        "self-alignment-interface",
        "text-interface",
    ],
    "SostenutoPedalLineSpanner": [
        "axis-group-interface",
        "grob-interface",
        "outside-staff-interface",
        "piano-pedal-interface",
        "side-position-interface",
        "spanner-interface",
    ],
    "SpacingSpanner": [
        "grob-interface",
        "spacing-options-interface",
        "spacing-spanner-interface",
        "spanner-interface",
    ],
    "SpanBar": [
        "bar-line-interface",
        "break-aligned-interface",
        "font-interface",
        "grob-interface",
        "item-interface",
        "span-bar-interface",
    ],
    "SpanBarStub": [
        "grob-interface",
        "item-interface",
        "pure-from-neighbor-interface",
    ],
    "StaffEllipsis": [
        "break-aligned-interface",
        "font-interface",
        "grob-interface",
        "item-interface",
        "text-interface",
    ],
    "StaffGrouper": [
        "grob-interface",
        "spanner-interface",
        "staff-grouper-interface",
    ],
    "StaffHighlight": [
        "grob-interface",
        "spanner-interface",
        "staff-highlight-interface",
    ],
    "StaffSpacing": [
        "grob-interface",
        "item-interface",
        "spacing-interface",
        "staff-spacing-interface",
    ],
    "StaffSymbol": [
        "grob-interface",
        "spanner-interface",
        "staff-symbol-interface",
    ],
    "StanzaNumber": [
        "font-interface",
        "grob-interface",
        "item-interface",
        "side-position-interface",
        "stanza-number-interface",
        "text-interface",
    ],
    "Stem": [
        "grob-interface",
        "item-interface",
        "stem-interface",
    ],
    "StemStub": [
        "grob-interface",
        "item-interface",
    ],
    "StemTremolo": [
        "grob-interface",
        "item-interface",
        "self-alignment-interface",
        "stem-tremolo-interface",
    ],
    "StringNumber": [
        "font-interface",
        "grob-interface",
        "item-interface",
        "number-interface",
        "outside-staff-interface",
        "self-alignment-interface",
        "side-position-interface",
        "string-number-interface",
        "text-interface",
        "text-script-interface",
    ],
    "StrokeFinger": [
        "font-interface",
        "grob-interface",
        "item-interface",
        "outside-staff-interface",
        "self-alignment-interface",
        "side-position-interface",
        "stroke-finger-interface",
        "text-interface",
        "text-script-interface",
    ],
    "SustainPedal": [
        "font-interface",
        "grob-interface",
        "item-interface",
        "piano-pedal-interface",
        "piano-pedal-script-interface",
        "self-alignment-interface",
        "text-interface",
    ],
    "SustainPedalLineSpanner": [
        "axis-group-interface",
        "grob-interface",
        "outside-staff-interface",
        "piano-pedal-interface",
        "side-position-interface",
        "spanner-interface",
    ],
    "System": [
        "axis-group-interface",
        "grob-interface",
        "outside-staff-axis-group-interface",
        "spanner-interface",
        "system-interface",
    ],
    "SystemStartBar": [
        "grob-interface",
        "side-position-interface",
        "spanner-interface",
        "system-start-delimiter-interface",
    ],
    "SystemStartBrace": [
        "font-interface",
        "grob-interface",
        "side-position-interface",
        "spanner-interface",
        "system-start-delimiter-interface",
    ],
    "SystemStartBracket": [
        "font-interface",
        "grob-interface",
        "side-position-interface",
        "spanner-interface",
        "system-start-delimiter-interface",
    ],
    "SystemStartSquare": [
        "font-interface",
        "grob-interface",
        "side-position-interface",
        "spanner-interface",
        "system-start-delimiter-interface",
    ],
    "TabNoteHead": [
        "bend-interface",
        "font-interface",
        "grob-interface",
        "item-interface",
        "note-head-interface",
        "rhythmic-grob-interface",
        "rhythmic-head-interface",
        "staff-symbol-referencer-interface",
        "tab-note-head-interface",
        "text-interface",
    ],
    "TextMark": [
        "accidental-switch-interface",
        "break-alignable-interface",
        "font-interface",
        "grob-interface",
        "item-interface",
        "mark-interface",
        "outside-staff-interface",
        "self-alignment-interface",
        "side-position-interface",
        "text-interface",
        "text-mark-interface",
    ],
    "TextScript": [
        "accidental-switch-interface",
        "font-interface",
        "grob-interface",
        "instrument-specific-markup-interface",
        "item-interface",
        "outside-staff-interface",
        "self-alignment-interface",
        "side-position-interface",
        "text-interface",
        "text-script-interface",
    ],
    "TextSpanner": [
        "font-interface",
        "grob-interface",
        "horizontal-line-spanner-interface",
        "line-interface",
        "line-spanner-interface",
        "outside-staff-interface",
        "side-position-interface",
        "spanner-interface",
    ],
    "Tie": [
        "bezier-curve-interface",
        "grob-interface",
        "spanner-interface",
        "tie-interface",
    ],
    "TieColumn": [
        "grob-interface",
        "spanner-interface",
        "tie-column-interface",
    ],
    "TimeSignature": [
        "break-aligned-interface",
        "font-interface",
        "grob-interface",
        "item-interface",
        "pure-from-neighbor-interface",
        "time-signature-interface",
    ],
    "TrillPitchAccidental": [
        "accidental-interface",
        "accidental-switch-interface",
        "font-interface",
        "grob-interface",
        "inline-accidental-interface",
        "item-interface",
        "side-position-interface",
        "trill-pitch-accidental-interface",
    ],
    "TrillPitchGroup": [
        "axis-group-interface",
        "grob-interface",
        "item-interface",
        "side-position-interface",
    ],
    "TrillPitchHead": [
        "font-interface",
        "grob-interface",
        "item-interface",
        "ledgered-interface",
        "pitched-trill-interface",
        "rhythmic-head-interface",
        "staff-symbol-referencer-interface",
    ],
    "TrillPitchParentheses": [
        "font-interface",
        "grob-interface",
        "item-interface",
        "parentheses-interface",
        "pitched-trill-interface",
    ],
    "TrillSpanner": [
        "font-interface",
        "grob-interface",
        "horizontal-line-spanner-interface",
        "line-interface",
        "line-spanner-interface",
        "outside-staff-interface",
        "side-position-interface",
        "spanner-interface",
        "trill-spanner-interface",
    ],
    "TupletBracket": [
        "grob-interface",
        "line-interface",
        "outside-staff-interface",
        "spanner-interface",
        "tuplet-bracket-interface",
    ],
    "TupletNumber": [
        "font-interface",
        "grob-interface",
        "outside-staff-interface",
        "spanner-interface",
        "text-interface",
        "tuplet-number-interface",
    ],
    "UnaCordaPedal": [
        "font-interface",
        "grob-interface",
        "item-interface",
        "piano-pedal-script-interface",
        "self-alignment-interface",
        "text-interface",
    ],
    "UnaCordaPedalLineSpanner": [
        "axis-group-interface",
        "grob-interface",
        "outside-staff-interface",
        "piano-pedal-interface",
        "side-position-interface",
        "spanner-interface",
    ],
    "VaticanaLigature": [
        "font-interface",
        "grob-interface",
        "spanner-interface",
        "vaticana-ligature-interface",
    ],
    "VerticalAlignment": [
        "align-interface",
        "axis-group-interface",
        "grob-interface",
        "spanner-interface",
    ],
    "VerticalAxisGroup": [
        "axis-group-interface",
        "grob-interface",
        "hara-kiri-group-spanner-interface",
        "outside-staff-axis-group-interface",
        "spanner-interface",
    ],
    "VoiceFollower": [
        "grob-interface",
        "line-interface",
        "line-spanner-interface",
        "spanner-interface",
    ],
    "VoltaBracket": [
        "font-interface",
        "grob-interface",
        "horizontal-bracket-interface",
        "line-interface",
        "side-position-interface",
        "spanner-interface",
        "text-interface",
        "volta-bracket-interface",
        "volta-interface",
    ],
    "VoltaBracketSpanner": [
        "axis-group-interface",
        "grob-interface",
        "outside-staff-interface",
        "side-position-interface",
        "spanner-interface",
        "volta-interface",
    ],
    "VowelTransition": [
        "grob-interface",
        "horizontal-line-spanner-interface",
        "line-interface",
        "line-spanner-interface",
        "lyric-interface",
        "spanner-interface",
    ],
}


contextproperties = [
    "aDueText",
    "accidentalGrouping",
    "additionalBassStrings",
    "additionalPitchPrefix",
    "alignAboveContext",
    "alignBelowContext",
    "alterationGlyphs",
    "alternativeNumber",
    "alternativeNumberingStyle",
    "alternativeRestores",
    "associatedVoice",
    "associatedVoiceType",
    "autoAccidentals",
    "autoBeamCheck",
    "autoBeaming",
    "autoCautionaries",
    "barCheckSynchronize",
    "barExtraVelocity",
    "barNumberFormatter",
    "barNumberVisibility",
    "baseMoment",
    "beamExceptions",
    "beamHalfMeasure",
    "beatExtraVelocity",
    "beatStructure",
    "breathMarkType",
    "caesuraType",
    "caesuraTypeTransform",
    "centerBarNumbers",
    "chordChanges",
    "chordNameExceptions",
    "chordNameFunction",
    "chordNameLowercaseMinor",
    "chordNameSeparator",
    "chordNoteNamer",
    "chordPrefixSpacer",
    "chordRootNamer",
    "clefGlyph",
    "clefPosition",
    "clefTransposition",
    "clefTranspositionFormatter",
    "clefTranspositionStyle",
    "codaMarkFormatter",
    "completionBusy",
    "completionFactor",
    "completionUnit",
    "connectArpeggios",
    "countPercentRepeats",
    "createKeyOnClefChange",
    "createSpacing",
    "crescendoSpanner",
    "crescendoText",
    "cueClefGlyph",
    "cueClefPosition",
    "cueClefTransposition",
    "cueClefTranspositionFormatter",
    "cueClefTranspositionStyle",
    "currentBarNumber",
    "dalSegnoTextFormatter",
    "decrescendoSpanner",
    "decrescendoText",
    "defaultStrings",
    "doubleRepeatBarType",
    "doubleRepeatSegnoBarType",
    "doubleSlurs",
    "drumPitchTable",
    "drumStyleTable",
    "endAtSkip",
    "endRepeatBarType",
    "endRepeatSegnoBarType",
    "explicitClefVisibility",
    "explicitCueClefVisibility",
    "explicitKeySignatureVisibility",
    "extendersOverRests",
    "extraNatural",
    "figuredBassAlterationDirection",
    "figuredBassCenterContinuations",
    "figuredBassFormatter",
    "figuredBassLargeNumberAlignment",
    "figuredBassPlusDirection",
    "figuredBassPlusStrokedAlist",
    "finalFineTextVisibility",
    "fineBarType",
    "fineSegnoBarType",
    "fineStartRepeatSegnoBarType",
    "fineText",
    "fingeringOrientations",
    "firstClef",
    "followVoice",
    "fontSize",
    "forbidBreak",
    "forbidBreakBetweenBarLines",
    "forceClef",
    "fretLabels",
    "glissandoMap",
    "gridInterval",
    "handleNegativeFrets",
    "harmonicAccidentals",
    "harmonicDots",
    "highStringOne",
    "ignoreBarChecks",
    "ignoreBarNumberChecks",
    "ignoreFiguredBassRest",
    "ignoreMelismata",
    "implicitBassFigures",
    "includeGraceNotes",
    "initialTimeSignatureVisibility",
    "instrumentCueName",
    "instrumentEqualizer",
    "instrumentName",
    "instrumentTransposition",
    "internalBarNumber",
    "keepAliveInterfaces",
    "keyAlterationOrder",
    "keyAlterations",
    "lyricMelismaAlignment",
    "lyricRepeatCountFormatter",
    "magnifyStaffValue",
    "majorSevenSymbol",
    "maximumFretStretch",
    "measureBarType",
    "measureLength",
    "measurePosition",
    "measureStartNow",
    "melismaBusyProperties",
    "metronomeMarkFormatter",
    "middleCClefPosition",
    "middleCCuePosition",
    "middleCOffset",
    "middleCPosition",
    "midiBalance",
    "midiChannelMapping",
    "midiChorusLevel",
    "midiExpression",
    "midiInstrument",
    "midiMaximumVolume",
    "midiMergeUnisons",
    "midiMinimumVolume",
    "midiPanPosition",
    "midiReverbLevel",
    "minimumFret",
    "minimumPageTurnLength",
    "minimumRepeatLengthForPageTurn",
    "minorChordModifier",
    "noChordSymbol",
    "noteNameFunction",
    "noteNameSeparator",
    "noteToFretFunction",
    "nullAccidentals",
    "ottavaStartNow",
    "ottavation",
    "ottavationMarkups",
    "output",
    "partCombineForced",
    "partCombineTextsOnNote",
    "pedalSostenutoStrings",
    "pedalSostenutoStyle",
    "pedalSustainStrings",
    "pedalSustainStyle",
    "pedalUnaCordaStrings",
    "pedalUnaCordaStyle",
    "predefinedDiagramTable",
    "printAccidentalNames",
    "printKeyCancellation",
    "printNotesLanguage",
    "printOctaveNames",
    "printPartCombineTexts",
    "proportionalNotationDuration",
    "rehearsalMark",
    "rehearsalMarkFormatter",
    "repeatCommands",
    "repeatCountVisibility",
    "restCompletionBusy",
    "restNumberThreshold",
    "restrainOpenStrings",
    "searchForVoice",
    "sectionBarType",
    "segnoBarType",
    "segnoMarkFormatter",
    "segnoStyle",
    "shapeNoteStyles",
    "shortInstrumentName",
    "shortVocalName",
    "skipBars",
    "skipTypesetting",
    "slashChordSeparator",
    "soloIIText",
    "soloText",
    "squashedPosition",
    "staffLineLayoutFunction",
    "stanza",
    "startAtNoteColumn",
    "startAtSkip",
    "startRepeatBarType",
    "startRepeatSegnoBarType",
    "stemLeftBeamCount",
    "stemRightBeamCount",
    "strictBeatBeaming",
    "stringNumberOrientations",
    "stringOneTopmost",
    "stringTunings",
    "strokeFingerOrientations",
    "subdivideBeams",
    "suggestAccidentals",
    "supportNonIntegerFret",
    "suspendMelodyDecisions",
    "suspendRestMerging",
    "systemStartDelimiter",
    "systemStartDelimiterHierarchy",
    "tabStaffLineLayoutFunction",
    "tablatureFormat",
    "tempoHideNote",
    "tempoWholesPerMinute",
    "tieWaitForNote",
    "timeSignatureFraction",
    "timeSignatureSettings",
    "timing",
    "tonic",
    "topLevelAlignment",
    "tupletFullLength",
    "tupletFullLengthNote",
    "tupletSpannerDuration",
    "underlyingRepeatBarType",
    "useBassFigureExtenders",
    "vocalName",
    "voltaSpannerDuration",
    "whichBar",
]


engravers = [
    "Accidental_engraver",
    "Alteration_glyph_engraver",
    "Ambitus_engraver",
    "Arpeggio_engraver",
    "Auto_beam_engraver",
    "Axis_group_engraver",
    "Balloon_engraver",
    "Bar_engraver",
    "Bar_number_engraver",
    "Beam_collision_engraver",
    "Beam_engraver",
    "Beam_performer",
    "Beat_engraver",
    "Beat_performer",
    "Bend_engraver",
    "Bend_spanner_engraver",
    "Break_align_engraver",
    "Breathing_sign_engraver",
    "Caesura_engraver",
    "Centered_bar_number_align_engraver",
    "Chord_name_engraver",
    "Chord_square_engraver",
    "Chord_tremolo_engraver",
    "Clef_engraver",
    "Cluster_spanner_engraver",
    "Collision_engraver",
    "Completion_heads_engraver",
    "Completion_rest_engraver",
    "Concurrent_hairpin_engraver",
    "Control_track_performer",
    "Cue_clef_engraver",
    "Current_chord_text_engraver",
    "Custos_engraver",
    "Divisio_engraver",
    "Dot_column_engraver",
    "Dots_engraver",
    "Double_percent_repeat_engraver",
    "Drum_note_performer",
    "Drum_notes_engraver",
    "Duration_line_engraver",
    "Dynamic_align_engraver",
    "Dynamic_engraver",
    "Dynamic_performer",
    "Episema_engraver",
    "Extender_engraver",
    "Figured_bass_engraver",
    "Figured_bass_position_engraver",
    "Finger_glide_engraver",
    "Fingering_column_engraver",
    "Fingering_engraver",
    "Font_size_engraver",
    "Footnote_engraver",
    "Forbid_line_break_engraver",
    "Fretboard_engraver",
    "Glissando_engraver",
    "Grace_auto_beam_engraver",
    "Grace_beam_engraver",
    "Grace_engraver",
    "Grace_spacing_engraver",
    "Grid_chord_name_engraver",
    "Grid_line_span_engraver",
    "Grid_point_engraver",
    "Grob_pq_engraver",
    "Horizontal_bracket_engraver",
    "Hyphen_engraver",
    "Instrument_name_engraver",
    "Instrument_switch_engraver",
    "Jump_engraver",
    "Keep_alive_together_engraver",
    "Key_engraver",
    "Key_performer",
    "Kievan_ligature_engraver",
    "Laissez_vibrer_engraver",
    "Ledger_line_engraver",
    "Ligature_bracket_engraver",
    "Lyric_engraver",
    "Lyric_performer",
    "Lyric_repeat_count_engraver",
    "Mark_engraver",
    "Mark_performer",
    "Mark_tracking_translator",
    "Measure_counter_engraver",
    "Measure_grouping_engraver",
    "Measure_spanner_engraver",
    "Melody_engraver",
    "Mensural_ligature_engraver",
    "Merge_mmrest_numbers_engraver",
    "Merge_rests_engraver",
    "Metronome_mark_engraver",
    "Midi_control_change_performer",
    "Multi_measure_rest_engraver",
    "New_fingering_engraver",
    "Non_musical_script_column_engraver",
    "Note_head_line_engraver",
    "Note_heads_engraver",
    "Note_name_engraver",
    "Note_performer",
    "Note_spacing_engraver",
    "Ottava_spanner_engraver",
    "Output_property_engraver",
    "Page_turn_engraver",
    "Paper_column_engraver",
    "Parenthesis_engraver",
    "Part_combine_engraver",
    "Percent_repeat_engraver",
    "Phrasing_slur_engraver",
    "Piano_pedal_align_engraver",
    "Piano_pedal_engraver",
    "Piano_pedal_performer",
    "Pitch_squash_engraver",
    "Pitched_trill_engraver",
    "Pure_from_neighbor_engraver",
    "Repeat_acknowledge_engraver",
    "Repeat_tie_engraver",
    "Rest_collision_engraver",
    "Rest_engraver",
    "Rhythmic_column_engraver",
    "Script_column_engraver",
    "Script_engraver",
    "Script_row_engraver",
    "Separating_line_group_engraver",
    "Show_control_points_engraver",
    "Signum_repetitionis_engraver",
    "Skip_typesetting_engraver",
    "Slash_repeat_engraver",
    "Slur_engraver",
    "Slur_performer",
    "Spacing_engraver",
    "Span_arpeggio_engraver",
    "Span_bar_engraver",
    "Span_bar_stub_engraver",
    "Span_stem_engraver",
    "Spanner_break_forbid_engraver",
    "Spanner_tracking_engraver",
    "Staff_collecting_engraver",
    "Staff_highlight_engraver",
    "Staff_performer",
    "Staff_symbol_engraver",
    "Stanza_number_align_engraver",
    "Stanza_number_engraver",
    "Stem_engraver",
    "System_start_delimiter_engraver",
    "Tab_note_heads_engraver",
    "Tab_staff_symbol_engraver",
    "Tab_tie_follow_engraver",
    "Tempo_performer",
    "Text_engraver",
    "Text_mark_engraver",
    "Text_spanner_engraver",
    "Tie_engraver",
    "Tie_performer",
    "Time_signature_engraver",
    "Time_signature_performer",
    "Timing_translator",
    "Trill_spanner_engraver",
    "Tuplet_engraver",
    "Tweak_engraver",
    "Vaticana_ligature_engraver",
    "Vertical_align_engraver",
    "Volta_engraver",
]


musicglyphs = [
    "accidentals.doublesharp",
    "accidentals.doublesharp.figbass",
    "accidentals.flat",
    "accidentals.flat.arrowboth",
    "accidentals.flat.arrowdown",
    "accidentals.flat.arrowup",
    "accidentals.flat.figbass",
    "accidentals.flat.koron",
    "accidentals.flat.slash",
    "accidentals.flat.slashslash",
    "accidentals.flatflat",
    "accidentals.flatflat.figbass",
    "accidentals.flatflat.slash",
    "accidentals.hufnagelM1",
    "accidentals.kievan1",
    "accidentals.kievanM1",
    "accidentals.leftparen",
    "accidentals.medicaeaM1",
    "accidentals.mensural1",
    "accidentals.mensuralM1",
    "accidentals.mirroredflat",
    "accidentals.mirroredflat.backslash",
    "accidentals.mirroredflat.flat",
    "accidentals.natural",
    "accidentals.natural.arrowboth",
    "accidentals.natural.arrowdown",
    "accidentals.natural.arrowup",
    "accidentals.natural.figbass",
    "accidentals.rightparen",
    "accidentals.sharp",
    "accidentals.sharp.arrowboth",
    "accidentals.sharp.arrowdown",
    "accidentals.sharp.arrowup",
    "accidentals.sharp.figbass",
    "accidentals.sharp.slash.stem",
    "accidentals.sharp.slashslash.stem",
    "accidentals.sharp.slashslash.stemstemstem",
    "accidentals.sharp.slashslashslash.stem",
    "accidentals.sharp.slashslashslash.stemstem",
    "accidentals.sharp.sori",
    "accidentals.vaticana0",
    "accidentals.vaticanaM1",
    "accordion.bayanbass",
    "accordion.discant",
    "accordion.dot",
    "accordion.freebass",
    "accordion.oldEE",
    "accordion.pull",
    "accordion.push",
    "accordion.stdbass",
    "arrowheads.close.01",
    "arrowheads.close.0M1",
    "arrowheads.close.11",
    "arrowheads.close.1M1",
    "arrowheads.open.01",
    "arrowheads.open.0M1",
    "arrowheads.open.11",
    "arrowheads.open.1M1",
    "backslash",
    "brackettips.down",
    "brackettips.up",
    "clefs.C",
    "clefs.C_change",
    "clefs.F",
    "clefs.F_change",
    "clefs.G",
    "clefs.GG",
    "clefs.GG_change",
    "clefs.G_change",
    "clefs.blackmensural.c",
    "clefs.blackmensural.c_change",
    "clefs.hufnagel.do",
    "clefs.hufnagel.do.fa",
    "clefs.hufnagel.do.fa_change",
    "clefs.hufnagel.do_change",
    "clefs.hufnagel.fa",
    "clefs.hufnagel.fa_change",
    "clefs.kievan.do",
    "clefs.kievan.do_change",
    "clefs.medicaea.do",
    "clefs.medicaea.do_change",
    "clefs.medicaea.fa",
    "clefs.medicaea.fa_change",
    "clefs.mensural.c",
    "clefs.mensural.c_change",
    "clefs.mensural.f",
    "clefs.mensural.f_change",
    "clefs.mensural.g",
    "clefs.mensural.g_change",
    "clefs.neomensural.c",
    "clefs.neomensural.c_change",
    "clefs.percussion",
    "clefs.percussion_change",
    "clefs.petrucci.c1",
    "clefs.petrucci.c1_change",
    "clefs.petrucci.c2",
    "clefs.petrucci.c2_change",
    "clefs.petrucci.c3",
    "clefs.petrucci.c3_change",
    "clefs.petrucci.c4",
    "clefs.petrucci.c4_change",
    "clefs.petrucci.c5",
    "clefs.petrucci.c5_change",
    "clefs.petrucci.f",
    "clefs.petrucci.f_change",
    "clefs.petrucci.g",
    "clefs.petrucci.g_change",
    "clefs.tab",
    "clefs.tab_change",
    "clefs.tenorG",
    "clefs.tenorG_change",
    "clefs.varC",
    "clefs.varC_change",
    "clefs.varpercussion",
    "clefs.varpercussion_change",
    "clefs.vaticana.do",
    "clefs.vaticana.do_change",
    "clefs.vaticana.fa",
    "clefs.vaticana.fa_change",
    "comma",
    "custodes.hufnagel.d0",
    "custodes.hufnagel.d1",
    "custodes.hufnagel.d2",
    "custodes.hufnagel.u0",
    "custodes.hufnagel.u1",
    "custodes.hufnagel.u2",
    "custodes.medicaea.d0",
    "custodes.medicaea.d1",
    "custodes.medicaea.d2",
    "custodes.medicaea.u0",
    "custodes.medicaea.u1",
    "custodes.medicaea.u2",
    "custodes.mensural.d0",
    "custodes.mensural.d1",
    "custodes.mensural.d2",
    "custodes.mensural.u0",
    "custodes.mensural.u1",
    "custodes.mensural.u2",
    "custodes.vaticana.d0",
    "custodes.vaticana.d1",
    "custodes.vaticana.d2",
    "custodes.vaticana.u0",
    "custodes.vaticana.u1",
    "custodes.vaticana.u2",
    "dots.dot",
    "dots.dotkievan",
    "dots.dotvaticana",
    "eight",
    "f",
    "fattened.eight",
    "fattened.five",
    "fattened.fixedwidth.eight",
    "fattened.fixedwidth.five",
    "fattened.fixedwidth.four",
    "fattened.fixedwidth.four.alt",
    "fattened.fixedwidth.nine",
    "fattened.fixedwidth.one",
    "fattened.fixedwidth.seven",
    "fattened.fixedwidth.seven.alt",
    "fattened.fixedwidth.six",
    "fattened.fixedwidth.three",
    "fattened.fixedwidth.two",
    "fattened.fixedwidth.zero",
    "fattened.four",
    "fattened.four.alt",
    "fattened.nine",
    "fattened.one",
    "fattened.seven",
    "fattened.seven.alt",
    "fattened.six",
    "fattened.three",
    "fattened.two",
    "fattened.zero",
    "figbass.fiveplus",
    "figbass.fourplus",
    "figbass.ninestroked",
    "figbass.sevenstroked",
    "figbass.sixstroked",
    "figbass.twoplus",
    "five",
    "fixedwidth.eight",
    "fixedwidth.five",
    "fixedwidth.four",
    "fixedwidth.four.alt",
    "fixedwidth.nine",
    "fixedwidth.one",
    "fixedwidth.seven",
    "fixedwidth.seven.alt",
    "fixedwidth.six",
    "fixedwidth.three",
    "fixedwidth.two",
    "fixedwidth.zero",
    "flags.d10",
    "flags.d3",
    "flags.d4",
    "flags.d5",
    "flags.d6",
    "flags.d7",
    "flags.d8",
    "flags.d9",
    "flags.dgrace",
    "flags.mensurald03",
    "flags.mensurald04",
    "flags.mensurald05",
    "flags.mensurald06",
    "flags.mensurald13",
    "flags.mensurald14",
    "flags.mensurald15",
    "flags.mensurald16",
    "flags.mensurald23",
    "flags.mensurald24",
    "flags.mensurald25",
    "flags.mensurald26",
    "flags.mensuralu03",
    "flags.mensuralu04",
    "flags.mensuralu05",
    "flags.mensuralu06",
    "flags.mensuralu13",
    "flags.mensuralu14",
    "flags.mensuralu15",
    "flags.mensuralu16",
    "flags.mensuralu23",
    "flags.mensuralu24",
    "flags.mensuralu25",
    "flags.mensuralu26",
    "flags.u10",
    "flags.u3",
    "flags.u4",
    "flags.u5",
    "flags.u6",
    "flags.u7",
    "flags.u8",
    "flags.u9",
    "flags.ugrace",
    "four",
    "four.alt",
    "hyphen",
    "m",
    "n",
    "nine",
    "noteheads.d0doFunk",
    "noteheads.d0fa",
    "noteheads.d0faFunk",
    "noteheads.d0faThin",
    "noteheads.d0miFunk",
    "noteheads.d0reFunk",
    "noteheads.d0tiFunk",
    "noteheads.d1doFunk",
    "noteheads.d1doWalker",
    "noteheads.d1fa",
    "noteheads.d1faFunk",
    "noteheads.d1faThin",
    "noteheads.d1faWalker",
    "noteheads.d1miFunk",
    "noteheads.d1reFunk",
    "noteheads.d1reWalker",
    "noteheads.d1tiFunk",
    "noteheads.d1tiWalker",
    "noteheads.d2doFunk",
    "noteheads.d2doWalker",
    "noteheads.d2fa",
    "noteheads.d2faFunk",
    "noteheads.d2faThin",
    "noteheads.d2faWalker",
    "noteheads.d2kievan",
    "noteheads.d2reFunk",
    "noteheads.d2reWalker",
    "noteheads.d2tiFunk",
    "noteheads.d2tiWalker",
    "noteheads.d3kievan",
    "noteheads.dM2",
    "noteheads.dM2blackmensural",
    "noteheads.dM2mensural",
    "noteheads.dM2neomensural",
    "noteheads.dM2semimensural",
    "noteheads.dM3blackmensural",
    "noteheads.dM3mensural",
    "noteheads.dM3neomensural",
    "noteheads.dM3semimensural",
    "noteheads.drM2mensural",
    "noteheads.drM2neomensural",
    "noteheads.drM2semimensural",
    "noteheads.drM3mensural",
    "noteheads.drM3neomensural",
    "noteheads.drM3semimensural",
    "noteheads.s0",
    "noteheads.s0blackmensural",
    "noteheads.s0blackpetrucci",
    "noteheads.s0cross",
    "noteheads.s0diamond",
    "noteheads.s0do",
    "noteheads.s0doThin",
    "noteheads.s0doWalker",
    "noteheads.s0faWalker",
    "noteheads.s0harmonic",
    "noteheads.s0kievan",
    "noteheads.s0la",
    "noteheads.s0laFunk",
    "noteheads.s0laThin",
    "noteheads.s0laWalker",
    "noteheads.s0mensural",
    "noteheads.s0mi",
    "noteheads.s0miMirror",
    "noteheads.s0miThin",
    "noteheads.s0miWalker",
    "noteheads.s0neomensural",
    "noteheads.s0petrucci",
    "noteheads.s0re",
    "noteheads.s0reThin",
    "noteheads.s0reWalker",
    "noteheads.s0slash",
    "noteheads.s0sol",
    "noteheads.s0solFunk",
    "noteheads.s0ti",
    "noteheads.s0tiThin",
    "noteheads.s0tiWalker",
    "noteheads.s0triangle",
    "noteheads.s1",
    "noteheads.s1blackpetrucci",
    "noteheads.s1cross",
    "noteheads.s1diamond",
    "noteheads.s1do",
    "noteheads.s1doThin",
    "noteheads.s1kievan",
    "noteheads.s1la",
    "noteheads.s1laFunk",
    "noteheads.s1laThin",
    "noteheads.s1laWalker",
    "noteheads.s1mensural",
    "noteheads.s1mi",
    "noteheads.s1miMirror",
    "noteheads.s1miThin",
    "noteheads.s1miWalker",
    "noteheads.s1neomensural",
    "noteheads.s1petrucci",
    "noteheads.s1re",
    "noteheads.s1reThin",
    "noteheads.s1slash",
    "noteheads.s1sol",
    "noteheads.s1solFunk",
    "noteheads.s1ti",
    "noteheads.s1tiThin",
    "noteheads.s1triangle",
    "noteheads.s2",
    "noteheads.s2blackpetrucci",
    "noteheads.s2cross",
    "noteheads.s2diamond",
    "noteheads.s2do",
    "noteheads.s2doThin",
    "noteheads.s2harmonic",
    "noteheads.s2la",
    "noteheads.s2laFunk",
    "noteheads.s2laThin",
    "noteheads.s2laWalker",
    "noteheads.s2mensural",
    "noteheads.s2mi",
    "noteheads.s2miFunk",
    "noteheads.s2miMirror",
    "noteheads.s2miThin",
    "noteheads.s2miWalker",
    "noteheads.s2neomensural",
    "noteheads.s2petrucci",
    "noteheads.s2re",
    "noteheads.s2reThin",
    "noteheads.s2slash",
    "noteheads.s2sol",
    "noteheads.s2solFunk",
    "noteheads.s2ti",
    "noteheads.s2tiThin",
    "noteheads.s2triangle",
    "noteheads.s2xcircle",
    "noteheads.sM1",
    "noteheads.sM1blackmensural",
    "noteheads.sM1double",
    "noteheads.sM1kievan",
    "noteheads.sM1mensural",
    "noteheads.sM1neomensural",
    "noteheads.sM1semimensural",
    "noteheads.sM2blackligmensural",
    "noteheads.sM2kievan",
    "noteheads.sM2ligmensural",
    "noteheads.sM2semiligmensural",
    "noteheads.sM3blackligmensural",
    "noteheads.sM3ligmensural",
    "noteheads.sM3semiligmensural",
    "noteheads.shufnagel.lpes",
    "noteheads.shufnagel.punctum",
    "noteheads.shufnagel.virga",
    "noteheads.smedicaea.inclinatum",
    "noteheads.smedicaea.punctum",
    "noteheads.smedicaea.rvirga",
    "noteheads.smedicaea.virga",
    "noteheads.sr1kievan",
    "noteheads.srM1mensural",
    "noteheads.srM1neomensural",
    "noteheads.srM1semimensural",
    "noteheads.srM2ligmensural",
    "noteheads.srM2semiligmensural",
    "noteheads.srM3ligmensural",
    "noteheads.srM3semiligmensural",
    "noteheads.ssolesmes.auct.asc",
    "noteheads.ssolesmes.auct.desc",
    "noteheads.ssolesmes.incl.auctum",
    "noteheads.ssolesmes.incl.parvum",
    "noteheads.ssolesmes.oriscus",
    "noteheads.ssolesmes.stropha",
    "noteheads.ssolesmes.stropha.aucta",
    "noteheads.svaticana.cephalicus",
    "noteheads.svaticana.epiphonus",
    "noteheads.svaticana.inclinatum",
    "noteheads.svaticana.inner.cephalicus",
    "noteheads.svaticana.linea.punctum",
    "noteheads.svaticana.linea.punctum.cavum",
    "noteheads.svaticana.lpes",
    "noteheads.svaticana.plica",
    "noteheads.svaticana.punctum",
    "noteheads.svaticana.punctum.cavum",
    "noteheads.svaticana.quilisma",
    "noteheads.svaticana.reverse.plica",
    "noteheads.svaticana.reverse.vplica",
    "noteheads.svaticana.upes",
    "noteheads.svaticana.vepiphonus",
    "noteheads.svaticana.vlpes",
    "noteheads.svaticana.vplica",
    "noteheads.svaticana.vupes",
    "noteheads.u0doFunk",
    "noteheads.u0fa",
    "noteheads.u0faFunk",
    "noteheads.u0faThin",
    "noteheads.u0miFunk",
    "noteheads.u0reFunk",
    "noteheads.u0tiFunk",
    "noteheads.u1doFunk",
    "noteheads.u1doWalker",
    "noteheads.u1fa",
    "noteheads.u1faFunk",
    "noteheads.u1faThin",
    "noteheads.u1faWalker",
    "noteheads.u1miFunk",
    "noteheads.u1reFunk",
    "noteheads.u1reWalker",
    "noteheads.u1tiFunk",
    "noteheads.u1tiWalker",
    "noteheads.u2doFunk",
    "noteheads.u2doWalker",
    "noteheads.u2fa",
    "noteheads.u2faFunk",
    "noteheads.u2faThin",
    "noteheads.u2faWalker",
    "noteheads.u2kievan",
    "noteheads.u2reFunk",
    "noteheads.u2reWalker",
    "noteheads.u2tiFunk",
    "noteheads.u2tiWalker",
    "noteheads.u3kievan",
    "noteheads.uM2",
    "noteheads.uM2blackmensural",
    "noteheads.uM2mensural",
    "noteheads.uM2neomensural",
    "noteheads.uM2semimensural",
    "noteheads.uM3blackmensural",
    "noteheads.uM3mensural",
    "noteheads.uM3neomensural",
    "noteheads.uM3semimensural",
    "noteheads.urM2mensural",
    "noteheads.urM2neomensural",
    "noteheads.urM2semimensural",
    "noteheads.urM3mensural",
    "noteheads.urM3neomensural",
    "noteheads.urM3semimensural",
    "one",
    "p",
    "pedal.*",
    "pedal..",
    "pedal.M",
    "pedal.P",
    "pedal.Ped",
    "pedal.d",
    "pedal.e",
    "period",
    "plus",
    "r",
    "rests.0",
    "rests.0mensural",
    "rests.0neomensural",
    "rests.0o",
    "rests.1",
    "rests.10",
    "rests.1mensural",
    "rests.1neomensural",
    "rests.1o",
    "rests.2",
    "rests.2classical",
    "rests.2mensural",
    "rests.2neomensural",
    "rests.2z",
    "rests.3",
    "rests.3mensural",
    "rests.3neomensural",
    "rests.4",
    "rests.4mensural",
    "rests.4neomensural",
    "rests.5",
    "rests.6",
    "rests.7",
    "rests.8",
    "rests.9",
    "rests.M1",
    "rests.M1mensural",
    "rests.M1neomensural",
    "rests.M1o",
    "rests.M2",
    "rests.M2mensural",
    "rests.M2neomensural",
    "rests.M3",
    "rests.M3mensural",
    "rests.M3neomensural",
    "s",
    "scripts.arpeggio",
    "scripts.arpeggio.arrow.1",
    "scripts.arpeggio.arrow.M1",
    "scripts.barline.kievan",
    "scripts.caesura.curved",
    "scripts.caesura.straight",
    "scripts.circulus",
    "scripts.coda",
    "scripts.daccentus",
    "scripts.dfermata",
    "scripts.dhenzelongfermata",
    "scripts.dhenzeshortfermata",
    "scripts.dlongfermata",
    "scripts.dmarcato",
    "scripts.downbow",
    "scripts.downmordent",
    "scripts.downprall",
    "scripts.dpedalheel",
    "scripts.dpedaltoe",
    "scripts.dportato",
    "scripts.dsemicirculus",
    "scripts.dshortfermata",
    "scripts.dsignumcongruentiae",
    "scripts.dstaccatissimo",
    "scripts.dverylongfermata",
    "scripts.dveryshortfermata",
    "scripts.espr",
    "scripts.flageolet",
    "scripts.halfopen",
    "scripts.halfopenvertical",
    "scripts.haydnturn",
    "scripts.ictus",
    "scripts.laltcomma",
    "scripts.lcomma",
    "scripts.lineprall",
    "scripts.lvarcomma",
    "scripts.mordent",
    "scripts.open",
    "scripts.prall",
    "scripts.pralldown",
    "scripts.prallmordent",
    "scripts.prallprall",
    "scripts.prallup",
    "scripts.raltcomma",
    "scripts.rcomma",
    "scripts.reverseturn",
    "scripts.rvarcomma",
    "scripts.segno",
    "scripts.sforzato",
    "scripts.slashturn",
    "scripts.snappizzicato",
    "scripts.staccato",
    "scripts.stopped",
    "scripts.tenuto",
    "scripts.thumb",
    "scripts.tickmark",
    "scripts.trill",
    "scripts.trill_element",
    "scripts.turn",
    "scripts.uaccentus",
    "scripts.ufermata",
    "scripts.uhenzelongfermata",
    "scripts.uhenzeshortfermata",
    "scripts.ulongfermata",
    "scripts.umarcato",
    "scripts.upbow",
    "scripts.upedalheel",
    "scripts.upedaltoe",
    "scripts.upmordent",
    "scripts.uportato",
    "scripts.upprall",
    "scripts.usemicirculus",
    "scripts.ushortfermata",
    "scripts.usignumcongruentiae",
    "scripts.ustaccatissimo",
    "scripts.uverylongfermata",
    "scripts.uveryshortfermata",
    "scripts.varcoda",
    "scripts.varsegno",
    "seven",
    "seven.alt",
    "six",
    "space",
    "three",
    "ties.lyric.default",
    "ties.lyric.short",
    "timesig.C22",
    "timesig.C44",
    "timesig.mensural22",
    "timesig.mensural24",
    "timesig.mensural32",
    "timesig.mensural34",
    "timesig.mensural44",
    "timesig.mensural48",
    "timesig.mensural64",
    "timesig.mensural68",
    "timesig.mensural68alt",
    "timesig.mensural94",
    "timesig.mensural98",
    "timesig.neomensural22",
    "timesig.neomensural24",
    "timesig.neomensural32",
    "timesig.neomensural34",
    "timesig.neomensural44",
    "timesig.neomensural48",
    "timesig.neomensural64",
    "timesig.neomensural68",
    "timesig.neomensural68alt",
    "timesig.neomensural94",
    "timesig.neomensural98",
    "two",
    "z",
    "zero",
]





# This file is part of python-ly, https://pypi.python.org/pypi/python-ly
#
# Copyright (c) 2008 - 2015 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
LilyPond information and logic concerning durations
"""

from __future__ import unicode_literals


from fractions import Fraction


durations = [
    '\\maxima', '\\longa', '\\breve',
    '1', '2', '4', '8', '16', '32', '64', '128', '256', '512', '1024', '2048'
]


def tostring(dur, dots=0, factor=1):
    r"""Returns the LilyPond string representation of a given logarithmic duration.
    
    Supports values from -3 up to and including 11.
    -2 = '\longa', 0  = '1' (whole note), etc.
    
    Adds the number of dots (defaults to 0) and the fraction factor if given.
    
    """
    s = durations[dur + 3] + '.' * dots
    if factor != 1:
        s += '*{0}'.format(factor)
    return s


def base_scaling(tokens):
    """Return (base, scaling) as two Fractions for the list of tokens."""
    base = Fraction(8, 1 << durations.index(tokens[0]))
    scaling = Fraction(1)
    half = base
    for t in tokens[1:]:
        if t == '.':
            half /= 2
            base += half
        elif t.startswith('*'):
            scaling *= Fraction(t[1:])
    return base, scaling


def base_scaling_string(duration):
    """Return (base, scaling) as two Fractions for the specified string."""
    items = duration.split('*')
    dots = items[0].split('.')
    base = Fraction(8, 1 << durations.index(dots[0].strip()))
    scaling = Fraction(1)
    half = base
    for dot in dots[1:]:
        half /= 2
        base += half
    for f in items[1:]:
        scaling *= Fraction(f.strip())
    return base, scaling


def fraction(tokens):
    """Return the duration of the Duration tokens as a Fraction."""
    base, scaling = base_scaling(tokens)
    return base * scaling


def fraction_string(duration):
    """Return the duration of the specified string as a Fraction."""
    base, scaling = base_scaling_string(duration)
    return base * scaling


def format_fraction(value):
    """Format the value as "5/1" etc."""
    if value == 0:
        return "0"
    elif isinstance(value, Fraction):
        return "{0}/{1}".format(value.numerator, value.denominator)
    else:
        return "{0}/1".format(value)





# node.py -- Node is a list-like type to build tree structures with
#
# Copyright (c) 2008 - 2015 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""

The Node class.

(c) 2008-2011 Wilbert Berendsen
License: GPL.

This module contains the Node class that can be used as a simple DOM (Document
Object Model) for building a tree structure.

A Node has children with list-like access methods and keeps also a reference to
its parent. A Node can have one parent; appending a Node to another Node causes
it to be removed from its parent node (if any).

To iterate over the children of a Node::

    for n in node:
        do_something(n)

To get the list of children of a Node::

    children = list(node)

Of course you can get the children directly using::

    child = node[3]

You should inherit from Node to make meaningful tree node types, e.g. to add
custom attributes or multiple sub-types.

A WeakNode class is provided as well, which uses a weak reference to the parent,
so that no cyclic references are created which might improve garbage collection.

"""

import weakref


class Node(object):
    """A list-like class to build tree structures with."""
    __slots__ = ('__weakref__', '_parent', '_children')
    
    def __init__(self, parent=None):
        self._parent = None
        self._children = []
        if parent:
            parent.append(self)

    def _own(self, node):
        """(Internal) Remove node from its parent if any and make us parent."""
        parent = node.parent()
        if parent:
            parent.remove(node)
        node._set_parent(self)
    
    def _set_parent(self, node):
        """(Internal) Set the Node (or None) as our parent."""
        self._parent = node
        
    def parent(self):
        """The parent, or None if the node has no parent."""
        return self._parent

    def index(self, node):
        """Return the index of the given child node."""
        return self._children.index(node)

    def append(self, node):
        """Append a node to the current node.
        
        It will be reparented, that means it will be removed from it's former
        parent if it had one.
        
        """
        self._own(node)
        self._children.append(node)
        
    def extend(self, iterable):
        """Append every Node from iterable."""
        for node in iterable:
            self.append(node)
        
    def insert(self, index, node):
        """Insert a node at the specified index."""
        self._own(node)
        self._children.insert(index, node)
        
    def insert_before(self, other, node):
        """Insert a node before the other node."""
        i = self.index(other)
        self._own(node)
        self._children.insert(i, node)
        
    def remove(self, node):
        """Remove the given child node."""
        self._children.remove(node)
        node._set_parent(None)

    def __bool__(self):
        """We are always true."""
        return True
    
    __nonzero__ = __bool__      # py2 compat
    
    def __len__(self):
        """Return the number of children."""
        return len(self._children)

    def __getitem__(self, k):
        """Return child at index or children at slice."""
        return self._children[k]

    def __setitem__(self, k, obj):
        """Set child at index or children at slice."""
        old = self._children[k]
        if isinstance(k, slice):
            if k.step:
                # extended slice, number of items must be same
                self._children[k] = obj
                # if this succeeded it's OK
                new = self._children[k]
            else:
                new = tuple(obj)
                self._children[k] = new
            for node in old:
                node._set_parent(None)
            for node in new:
                self._own(node)
        else:
            old._set_parent(None)
            self._children[k] = obj
            self._own(obj)

    def __delitem__(self, k):
        """Delete child at index or children at slice."""
        if isinstance(k, slice):
            for node in self._children[k]:
                node._set_parent(None)
        else:
            self._children[k]._set_parent(None)
        del self._children[k]

    def __contains__(self, node):
        """Return True if the node is our child."""
        return node in self._children

    def clear(self):
        """Remove all children."""
        del self[:]

    def unlink(self):
        """Remove all children and unlink() them as well."""
        for node in self:
            node.unlink()
        del self._children[:]
    
    def replace(self, old, new):
        """Replace a child node with another node."""
        i = self.index(old)
        self[i] = new

    def sort(self, key=None, reverse=False):
        """Sorts the children, optionally using the key function.
        
        Using a key function is recommended, or you must add comparison methods
        to your Node subclass.
        
        """
        self._children.sort(key, reverse=reverse)
        
    def copy(self):
        """Return a deep copy of the node and its children """
        obj = self.__class__.__new__(self.__class__)
        obj._parent = None
        obj._children = []
        self._copy_attrs(obj)
        for n in self:
            obj.append(n.copy())
        return obj
    
    def _copy_attrs(self, node):
        """Called by copy(); copy attributes not starting with '_'."""
        for name, value in vars(self).items():
            name.startswith("_") or setattr(node, name, value)
            
    def ancestors(self):
        """Climb the tree up over the parents."""
        node = self.parent()
        while node:
            yield node
            node = node.parent()

    def previous_sibling(self):
        """Return the sibling object just before us in our parents list.
        
        Returns None if this is the first child, or if we have no parent.
        
        """
        for i in self.backward():
            return i

    def next_sibling(self):
        """Return the sibling object just after us in our parents list.
        
        Returns None if this is the last child, or if we have no parent.
        
        """
        for i in self.forward():
            return i

    def backward(self):
        """Iterate (backwards) over the preceding siblings."""
        parent = self.parent()
        if parent:
            i = parent.index(self)
            return iter(parent[i-1::-1])

    def forward(self):
        """Iterate over the following siblings."""
        parent = self.parent()
        if parent:
            i = parent.index(self)
            return iter(parent[i+1::])

    def is_descendant_of(self, parent):
        """Return True if self is a descendant of parent, else False."""
        for n in self.ancestors():
            if n is parent:
                return True
        return False

    def toplevel(self):
        """Return the toplevel parent Node of this node."""
        node = self
        parent = self.parent()
        while parent:
            node = parent
            parent = node.parent()
        return node
    
    def descendants(self, depth = -1):
        """Yield all the descendants, in tree order. Same as iter_depth()."""
        return self.iter_depth(depth)
    
    def iter_depth(self, depth = -1):
        """Iterate over all the children, and their children, etc.
        
        Set depth to restrict the search to a certain depth, -1 is unrestricted.
        
        """
        if depth != 0:
            for i in self:
                yield i
                for j in i.iter_depth(depth - 1):
                    yield j
                
    def iter_rings(self, depth = -1):
        """Iterate over the children in rings, depth last.
        
        This method returns the closest descendants first.
        Set depth to restrict the search to a certain depth, -1 is unrestricted.
        
        """
        children = list(self)
        while children and depth:
            depth -= 1
            newchildren = []
            for i in children:
                yield i
                newchildren.extend(i)
            children = newchildren

    def find(self, cls, depth = -1):
        """Yield all descendants if they are an instance of cls.
        
        cls may also be a tuple of classes. This method uses iter_depth().
        
        """
        for node in self.iter_depth(depth):
            if isinstance(node, cls):
                yield node
        
    def find_children(self, cls, depth = -1):
        """Yield all descendants if they are an instance of cls.
        
        cls may also be a tuple of classes. This method uses iter_rings().
        
        """
        for node in self.iter_rings(depth):
            if isinstance(node, cls):
                yield node

    def find_child(self, cls, depth = -1):
        """Return the first descendant that's an instance of cls.
        
        cls may also be a tuple of classes. This method uses iter_rings().
        
        """
        for node in self.iter_rings(depth):
            if isinstance(node, cls):
                return node
    
    def find_parent(self, cls):
        """Find an ancestor that's an instance of the given class.
        
        cls may also be a tuple of classes.
        
        """
        for node in self.ancestors():
            if isinstance(node, cls):
                return node
    
    def dump(self):
        """Return a string representation of the tree."""
        def line(obj, indent):
            yield indent * "  " + repr(obj)
            for c in obj:
                for l in line(c, indent + 1):
                    yield l
        return '\n'.join(line(self, 0))



class WeakNode(Node):
    """A Node type using a weak reference to the parent."""
    __slots__ = ()
    def _set_parent(self, node):
        self._parent = None if node is None else weakref.ref(node)
    
    def parent(self):
        """The parent, or None if the node has no parent."""
        if self._parent is not None:
            return self._parent()





# This file is part of python-ly, https://pypi.python.org/pypi/python-ly
#
# Copyright (c) 2014 - 2015 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
Options configuring the commands' behaviour
"""

from __future__ import unicode_literals

from ly.cli import setvar

class ServerOptions(object):
    """
    Options configuring the server itself, not the individual commands
    """
    
    def __init__(self):
        self.port = 5432
        self.timeout = 0
    

class Options(object):
    """Store all the startup options and their defaults."""
    def __init__(self):
        self.mode = None
        self.in_place = False
        self.encoding = 'UTF-8'
        self.output_encoding = None
        self.output = None
        self.replace_pattern = True
        self.backup_suffix = '~'
        self.with_filename = None
        self.default_language = "nederlands"
        
        self.indent_width = 2
        self.indent_tabs = False
        self.tab_width = 8
        
        self.full_html = True
        self.inline_style = False
        self.stylesheet = None
        self.number_lines = False
        self.wrapper_tag = 'pre'
        self.wrapper_attribute = 'class'
        self.document_id = 'lilypond'
        self.linenumbers_id = 'linenumbers'

    def set_variable(self, name, value):
        name = name.replace('-', '_')
        try:
            func = getattr(setvar, name)
        except AttributeError:
            raise AttributeError("unknown variable: {name}".format(name=name))
        try:
            value = func(value)
        except ValueError as e:
            raise ValueError(format(e))
        setattr(self, name, value)    
